/*
**   Generic Config file parser (configparse.c)
**   Copyright (C) 1999 Fred Jacobs, Chris Behrens
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
**  The idea of something this generic came solely from Fred Jacobs.
**  I completely rewrote his implementation, however, although some
**  pieces remain similar.  Thank you very much, Fred! :)
**    - Chris
*/

#include "ircd_includes.h"
#include "configparse.h"

typedef struct _config_info
{
	config_errfunc_t	err_func;
} CONFIG_INFO;

static CONFIGFILE *config_parse(CONFIG_INFO *cinfo, char *filename, char *confdata);
static void config_entry_free(CONFIGENTRY *ceptr);

static CONFIGFILE *config_parse(CONFIG_INFO *cinfo, char *filename, char *confdata)
{
	char		*ptr;
	char		*start;
	int			linenumber = 1;
	CONFIGENTRY	*curce;
	CONFIGENTRY	**lastce;
	CONFIGENTRY	*cursection;

	CONFIGFILE	*curcf;
	CONFIGFILE	*lastcf;

	lastcf = curcf = (CONFIGFILE *)MyMalloc(sizeof(CONFIGFILE));
	memset(curcf, 0, sizeof(CONFIGFILE));
	curcf->cf_filename = MyStrdup(filename);
	lastce = &(curcf->cf_entries);
	curce = NULL;
	cursection = NULL;

	ptr = confdata;
	while(*ptr == '#')
	{
		while(*++ptr && (*ptr != '\n'))
			;
		if (!*ptr)
		{
			break;
		}
		linenumber++;
		ptr++;
	}

	for(;*ptr;ptr++)
	{
		switch(*ptr)
		{
			case ';':
				if (!curce)
				{
					cinfo->err_func(LOG_NOTICE, "%s:%i Ignoring extra semicolon\n",
						filename, linenumber);
					break;
				}
				if (!strcmp(curce->ce_varname, "include"))
				{
					CONFIGFILE	*cfptr;

					if (!curce->ce_vardata)
					{
						cinfo->err_func(LOG_NOTICE, "%s:%i Ignoring \"include\": No filename given\n",
							filename, linenumber);
						config_entry_free(curce);
						curce = NULL;
						continue;
					}
					if (strlen(curce->ce_vardata) > 255)
						curce->ce_vardata[255] = '\0';
					cfptr = config_load(curce->ce_vardata, cinfo->err_func);
					if (cfptr)
					{
						lastcf->cf_next = cfptr;
						lastcf = cfptr;
					}
					config_entry_free(curce);
					curce = NULL;
					continue;
				}
				*lastce = curce;
				lastce = &(curce->ce_next);
				curce->ce_fileposend = (ptr - confdata);
				curce = NULL;
				break;
			case '{':
				if (!curce)
				{
					cinfo->err_func(LOG_NOTICE, "%s:%i: No name for section start\n",
							filename, linenumber);
					continue;
				}
				else if (curce->ce_entries)
				{
					cinfo->err_func(LOG_NOTICE, "%s:%i: Ignoring extra section start\n",
							filename, linenumber);
					continue;
				}
				curce->ce_sectlinenum = linenumber;
				lastce = &(curce->ce_entries);
				cursection = curce;
				curce = NULL;
				break;
			case '}':
				if (curce)
				{
					cinfo->err_func(LOG_ERR, "%s:%i: Missing semicolon before close brace\n",
						filename, linenumber);
					config_entry_free(curce);
					config_free(curcf);
					return NULL;
				}
				else if (!cursection)
				{
					cinfo->err_func(LOG_NOTICE, "%s:%i: Ignoring extra close brace\n",
						filename, linenumber);
					continue;
				}
				curce = cursection;
				cursection->ce_fileposend = (ptr - confdata);
				cursection = cursection->ce_prevlevel;
				if (!cursection)
					lastce = &(curcf->cf_entries);
				else
					lastce = &(cursection->ce_entries);
				for(;*lastce;lastce = &((*lastce)->ce_next))
					continue;
				break;
			case '/':
				if (*(ptr+1) == '/')
				{
					ptr += 2;
					while(*ptr && (*ptr != '\n'))
						ptr++;
					if (!*ptr)
						break;
					ptr--; /* grab the \n on next loop thru */
					continue;
				}
				else if (*(ptr+1) == '*')
				{
					int commentstart = linenumber;

					for(ptr+=2;*ptr;ptr++)
					{
						if ((*ptr == '*') && (*(ptr+1) == '/'))
						{
							ptr++;
							break;
						}
						else if (*ptr == '\n')
							linenumber++;
					}
					if (!*ptr)
					{
						cinfo->err_func(LOG_ERR, "%s:%i Comment on this line does not end\n",
							filename, commentstart);
						config_entry_free(curce);
						config_free(curcf);
						return NULL;
					}
				}
				break;
			case '\"':
				start = ++ptr;
				for(;*ptr;ptr++)
				{
					if ((*ptr == '\\') && (*(ptr+1) == '\"'))
					{
						char *tptr = ptr;
						while((*tptr = *(tptr+1)))
							tptr++;
					}
					else if ((*ptr == '\"') || (*ptr == '\n'))
						break;
				}
				if (!*ptr || (*ptr == '\n'))
				{
					cinfo->err_func(LOG_ERR, "%s:%i: Unterminated quote found\n",
							filename, linenumber);
					config_entry_free(curce);
					config_free(curcf);
					return NULL;
				}
				if (curce)
				{
					if (curce->ce_vardata)
					{
						cinfo->err_func(LOG_NOTICE, "%s:%i: Ignoring extra data\n",
							filename, linenumber);
					}
					else
					{
						char	*eptr;

						curce->ce_vardata = (char *)MyMalloc(ptr-start+1);
						strncpy(curce->ce_vardata, start, ptr-start);
						curce->ce_vardata[ptr-start] = '\0';
						curce->ce_vardatanum =
							strtol(curce->ce_vardata, &eptr, 0) &
								0xffffffff;	/* we only want 32bits and
									long is 64bit on 64bit compiles */
						if (eptr != (curce->ce_vardata + (ptr - start)))
						{
							curce->ce_vardatanum = 0;
						}
					}
				}
				else
				{
					curce = (CONFIGENTRY *)MyMalloc(sizeof(CONFIGENTRY));
					memset(curce, 0, sizeof(CONFIGENTRY));
					curce->ce_varname = (char *)MyMalloc(ptr-start+1);
					strncpy(curce->ce_varname, start, ptr-start);
					curce->ce_varname[ptr-start] = '\0';
					curce->ce_varlinenum = linenumber;
					curce->ce_fileptr = curcf;
					curce->ce_prevlevel = cursection;
					curce->ce_fileposstart = (start - confdata);
				}
				break;
			case '\n':
				linenumber++;
				while (*(ptr+1) == '#')
				{
					++ptr;
					while(*++ptr && (*ptr != '\n'))
						;
					if (!*ptr)
					{
						/* make for(;;) exit from the loop */
						ptr--;
						break; /* will break to the above while,
									but that check will fail */
					}
					linenumber++;
				}
				break;
			case '\t':
			case ' ':
			case '\r':
				break;
			default:
				if ((*ptr == '*') && (*(ptr+1) == '/'))
				{
					cinfo->err_func(LOG_NOTICE, "%s:%i Ignoring extra end comment\n",
						filename, linenumber);
					ptr++;
					break;
				}
				start = ptr;
				for(;*ptr;ptr++)
				{
					if ((*ptr == ' ') || (*ptr == '\t') || (*ptr == '\n') || (*ptr == ';'))
						break;
				}
				if (!*ptr)
				{
					if (curce)
						cinfo->err_func(LOG_ERR, "%s: Unexpected EOF for variable starting at %i\n",
							filename, curce->ce_varlinenum);
					else if (cursection)
						cinfo->err_func(LOG_ERR, "%s: Unexpected EOF for section starting at %i\n",
							filename, curce->ce_sectlinenum);
					else
						cinfo->err_func(LOG_ERR, "%s: Unexpected EOF.\n", filename);
					config_entry_free(curce);
					config_free(curcf);
					return NULL;
				}
				if (curce)
				{
					if (curce->ce_vardata)
					{
						cinfo->err_func(LOG_NOTICE, "%s:%i: Ignoring extra data\n",
							filename, linenumber);
					}
					else
					{
						char *eptr;

						curce->ce_vardata = (char *)MyMalloc(ptr-start+1);
						strncpy(curce->ce_vardata, start, ptr-start);
						curce->ce_vardata[ptr-start] = '\0';
						curce->ce_vardatanum =
							strtol(curce->ce_vardata, &eptr, 0) &
								0xffffffff;	/* we only want 32bits and
									long is 64bit on 64bit compiles */
						if (eptr != (curce->ce_vardata + (ptr - start)))
						{
							curce->ce_vardatanum = 0;
						}
					}
				}
				else
				{
					curce = (CONFIGENTRY *)MyMalloc(sizeof(CONFIGENTRY));
					memset(curce, 0, sizeof(CONFIGENTRY));
					curce->ce_varname = (char *)MyMalloc(ptr-start+1);
					strncpy(curce->ce_varname, start, ptr-start);
					curce->ce_varname[ptr-start] = '\0';
					curce->ce_varlinenum = linenumber;
					curce->ce_fileptr = curcf;
					curce->ce_prevlevel = cursection;
					curce->ce_fileposstart = (start - confdata);
				}
				if ((*ptr == ';') || (*ptr == '\n'))
					ptr--;
				break;
		} /* switch */
	} /* for */
	if (curce)
	{
		cinfo->err_func(LOG_ERR, "%s: Unexpected EOF for variable starting on line %i\n",
			filename, curce->ce_varlinenum);
		config_entry_free(curce);
		config_free(curcf);
		return NULL;
	}
	else if (cursection)
	{
		cinfo->err_func(LOG_ERR, "%s: Unexpected EOF for section starting on line %i\n",
				filename, cursection->ce_sectlinenum);
		config_free(curcf);
		return NULL;
	}
	return curcf;
}

static void config_entry_free(CONFIGENTRY *ceptr)
{
	CONFIGENTRY	*nptr;

	for(;ceptr;ceptr=nptr)
	{
		nptr = ceptr->ce_next;
		if (ceptr->ce_entries)
			config_entry_free(ceptr->ce_entries);
		if (ceptr->ce_varname)
			MyFree(ceptr->ce_varname);
		if (ceptr->ce_vardata)
			MyFree(ceptr->ce_vardata);
		MyFree(ceptr);
	}
}

void config_free(CONFIGFILE *cfptr)
{
	CONFIGFILE	*nptr;

	for(;cfptr;cfptr=nptr)
	{
		nptr = cfptr->cf_next;
		if (cfptr->cf_entries)
			config_entry_free(cfptr->cf_entries);
		if (cfptr->cf_filename)
			MyFree(cfptr->cf_filename);
		MyFree(cfptr);
	}
}

CONFIGFILE *config_load(char *filename, config_errfunc_t err_func)
{
	struct stat sb;
	int			fd;
	int			ret;
	char		*buf = NULL;
	CONFIGFILE	*cfptr;
	CONFIG_INFO	cinfobuf;
	CONFIG_INFO	*cinfo = &cinfobuf;

	memset(&cinfobuf, 0, sizeof(cinfobuf));
	cinfo->err_func = err_func ? err_func : (config_errfunc_t)syslog;

	fd = open(filename, O_RDONLY);
	if (fd == -1)
	{
		cinfo->err_func(LOG_ERR, "Couldn't open \"%s\": %s\n", filename, strerror(errno));
		return NULL;
	}
	if (fstat(fd, &sb) == -1)
	{
		cinfo->err_func(LOG_ERR, "Couldn't fstat \"%s\": %s\n", filename, strerror(errno));
		close(fd);
		return NULL;
	}
	if (!sb.st_size)
	{
		close(fd);
		return NULL;
	}
	buf = (char *)MyMalloc(sb.st_size+1);
	if (buf == NULL)
	{
		cinfo->err_func(LOG_ERR, "Out of memory trying to load \"%s\"\n", filename);
		close(fd);
		return NULL;
	}
	ret = read(fd, buf, sb.st_size);
	if (ret != sb.st_size)
	{
		cinfo->err_func(LOG_ERR, "Error reading \"%s\": %i\n", filename,
			ret == -1 ? strerror(errno) : strerror(EFAULT));
		MyFree(buf);
		close(fd);
		return NULL;
	}
	buf[ret] = '\0';
	close(fd);
	cfptr = config_parse(cinfo, filename, buf);
	MyFree(buf);
	return cfptr;
}

CONFIGENTRY *config_find(CONFIGENTRY *ceptr, char *name)
{
	for(;ceptr;ceptr=ceptr->ce_next)
		if (!strcmp(ceptr->ce_varname, name))
			break;
	return ceptr;
}

