/*
**   IRC - Internet Relay Chat, source/conf.c
**   Copyright (C) 1999 Chris Behrens
**
**   Portions Copyright (C) 1990 Jarkko Oikarinen and
**                      University of Oulu, Computing Center
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ircd_includes.h"
#include "conf.h"
#include "../confparse/configparse.h"
#include "commands.h"
#include "access.h"

RCSID("$Id: conf.c,v 1.37 2003/03/10 10:31:48 cbehrens Exp $");

#define IRCDCONF_DEFAULT_SENDQ			100000
#define IRCDCONF_DEFAULT_PINGFREQ		120
#define IRCDCONF_DEFAULT_CONNFREQ		300
#define IRCDCONF_DEFAULT_FLOODTIME		2
#define IRCDCONF_DEFAULT_FLOODNUM		4
#define IRCDCONF_DEFAULT_FLOODEXIT		0

typedef struct _confcommand		CONFCOMMAND;

struct _confcommand
{
	char	*name;
	int		(*func)(IRCDCONF *conf, CONFIGENTRY *ce, char *cmd);
};

static int _servlink_res_callback(IRCDRES *ircdres, IRCDRES_RESULT *result, SERVLINK *sl);
static int _conf_admin(IRCDCONF *conf, CONFIGENTRY *ce, char *cmd);
static int _conf_allow(IRCDCONF *conf, CONFIGENTRY *ce, char *cmd);
static int _conf_class(IRCDCONF *conf, CONFIGENTRY *ce, char *cmd);
static int _conf_deny(IRCDCONF *conf, CONFIGENTRY *ce, char *cmd);
static int _conf_link(IRCDCONF *conf, CONFIGENTRY *ce, char *cmd);
static int _conf_listen(IRCDCONF *conf, CONFIGENTRY *ce, char *cmd);
static int _conf_global(IRCDCONF *conf, CONFIGENTRY *ce, char *cmd);
static int _conf_opers(IRCDCONF *conf, CONFIGENTRY *ce, char *cmd);
static int _conf_set(IRCDCONF *conf, CONFIGENTRY *ce, char *cmd);
static void logs_flushall(void);
static void _isupport_add_num(IRCD_STRINGLIST **sl, char *buffer, u_int bufsize, char **curptr, char *variable, int num);
static void _isupport_add_string(IRCD_STRINGLIST **sl, char *buffer, u_int bufsize, char **curptr, char *variable, char *string);
static void _isupport_done(IRCD_STRINGLIST **sl, char *buffer);
static int _deny_acl_add(IRCD_DENY_ACL **acl_list, IRCD_DENY_ACL **acl_last, char *oper, char *server, char *mask, u_int flags, IRCD_DENY_ACL **aclret);
static void _deny_acls_free(IRCD_DENY_ACL **acl_list, IRCD_DENY_ACL **acl_last);
#if 0
static void logs_closeall(void);
#endif

static CONFCOMMAND _ConfCommands[] = {
	{	"ADMIN",		_conf_admin			},
	{	"ALLOW",		_conf_allow			},
	{	"CLASS",		_conf_class			},
	{	"DENY",			_conf_deny			},
	{	"LINK",			_conf_link			},
	{	"CONNECT",		_conf_link			},
	{	"SERVER",		_conf_link			},
	{	"LISTEN",		_conf_listen		},
	{	"GLOBAL",		_conf_global		},
	{	"OPERS",		_conf_opers			},
	{	"SET",			_conf_set			},
	{	NULL,			NULL				}
};


static int _servlink_res_callback(IRCDRES *ircdres, IRCDRES_RESULT *result, SERVLINK *sl)
{
	del_reqlist_from_list(&Reqlist, sl->res_req);
	MyFree(sl->res_req);
	sl->res_req = NULL;
	sl->flags &= ~SERVLINK_FLAGS_DOINGDNS;
	if (!result)
	{
		send_to_opers("Lookup of %s for \"link %s\" conf entry failed",
			sl->hostname, sl->servername);
	}
	else
	{
		if (result->res_he->he_he.h_length == sizeof(struct in_addr))
			BCOPY(result->res_he->he_he.h_addr_list[0], &sl->inaddr,
				sizeof(struct in_addr));
		MyFree(result->res_he);
	}
	return 0;
}

static int _conf_admin(IRCDCONF *conf, CONFIGENTRY *ce, char *cmd)
{
	char	**ptr;
	int		size = 5;
	int		num = 0;
	char	*line;

	if (!ce->ce_entries)
	{
		syslog(LOG_NOTICE, "%s:%i: no section for motd",
				ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		return -1;
	}

	if (conf->admin)
	{
		for(ptr=conf->admin;*ptr;ptr++)
			MyFree(*ptr);
		MyFree(conf->admin);
		conf->admin = NULL;
	}

	ptr = conf->admin = (char **)MyMalloc(sizeof(char *)*size);

	for(ce=ce->ce_entries;ce;ce=ce->ce_next)
	{
		if (!(line = ce->ce_varname))
			continue;
		if (++num == size)
		{
			int diff = ptr - conf->admin;

			size += 5;
			conf->admin = (char **)MyRealloc(conf->admin, sizeof(char *)*size);
			ptr = conf->admin + diff;
		}
		while(*line && ISWHITE(*line))
			line++;
		*ptr = (char *)MyStrdup(line);
		ptr++;
	}
	*ptr = NULL;
	return 0;
}

static int _conf_allow(IRCDCONF *conf, CONFIGENTRY *ce, char *cmd)
{
	char		*username = NULL;
	char		*hostname = NULL;
	char		*ipstring = NULL;
	char		*password = NULL;
	char		*gline_for_mask = NULL;
	char		*gline_from_oper = NULL;
	char		*gline_from_server = NULL;
	char		*kline_for_mask = NULL;
	char		*kline_from_oper = NULL;
	char		*kline_from_server = NULL;
	char		*unkline_for_mask = NULL;
	char		*unkline_from_oper = NULL;
	char		*unkline_from_server = NULL;
	int			flags = 0;
	char		class[MAXCLASSNAME+1];
	char		*comment = NULL;
	CONFIGENTRY	*ce2;
	int			linestart = ce->ce_sectlinenum;
	char		*filename = ce->ce_fileptr->cf_filename;
	int			num_commands = 0;

	if (!ce->ce_entries)
	{
		syslog(LOG_NOTICE, "%s:%i: no section for allow",
				ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		return -1;
	}
	*class = '\0';
	for(ce=ce->ce_entries;ce;ce=ce->ce_next)
	{
		if (!ce->ce_varname)
			continue;
		if (!strcasecmp(ce->ce_varname, "hostname"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no hostname given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!hostname)
				hostname = ce->ce_vardata;
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra hostname",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "username"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no username given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!username)
				username = ce->ce_vardata;
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra username",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "ipaddr"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no ipaddress given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!ipstring)
				ipstring = ce->ce_vardata;
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra ipaddress",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "password"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no password given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!password)
				password = ce->ce_vardata;
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra password",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "gline_for_mask"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no user@host given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!gline_for_mask)
				gline_for_mask = ce->ce_vardata;
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra 'gline_for_mask'",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "gline_from_oper"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no user@host given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!gline_from_oper)
				gline_from_oper = ce->ce_vardata;
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra 'gline_from_oper'",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "gline_from_server"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no server given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!gline_from_server)
				gline_from_server = ce->ce_vardata;
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra 'gline_from_server'",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "kline_for_mask"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no user@host given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!kline_for_mask)
				kline_for_mask = ce->ce_vardata;
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra 'kline_for_mask'",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "kline_from_oper"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no user@host given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!kline_from_oper)
				kline_from_oper = ce->ce_vardata;
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra 'kline_from_oper'",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "kline_from_server"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no server given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!kline_from_server)
				kline_from_server = ce->ce_vardata;
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra 'kline_from_server'",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "unkline_for_mask"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no user@host given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!unkline_for_mask)
				unkline_for_mask = ce->ce_vardata;
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra 'unkline_for_mask'",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "unkline_from_oper"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no user@host given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!unkline_from_oper)
				unkline_from_oper = ce->ce_vardata;
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra 'unkline_from_oper'",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "unkline_from_server"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no server given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!unkline_from_server)
				unkline_from_server = ce->ce_vardata;
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra 'unkline_from_server'",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "flag"))
		{
			ce2 = ce->ce_entries ? ce->ce_entries : ce;
			for(;ce2;ce2=(ce2 == ce) ? NULL : ce2->ce_next)
			{
				char *flagname = (ce2 == ce) ? ce2->ce_vardata : ce2->ce_varname;
				if (!flagname)
				{
					syslog(LOG_NOTICE, "%s:%i: no flag given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
					continue;
				}
				if (!strcasecmp(flagname, "checkhostname"))
					flags |= ALLOW_FLAGS_CHECKHOSTNAME;
				else
					syslog(LOG_NOTICE,
						"%s:%i: unknown flag",
						ce->ce_fileptr->cf_filename, ce2->ce_varlinenum);
			}
		}
		else if (!strcasecmp(ce->ce_varname, "class"))
		{
			if (!ce->ce_vardata || !*ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no class specified",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!*class)
			{
				support_strncpy(class, ce->ce_vardata, sizeof(class));
			}
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra class",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "comment"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no comment given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!comment)
				comment = ce->ce_vardata;
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra comment",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else
			syslog(LOG_NOTICE, "%s:%i: unknown option for allow",
				ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
	}

	if (hostname || ipstring)
		num_commands++;
	if (gline_for_mask || gline_from_oper || gline_from_server)
		num_commands++;
	if (kline_from_oper || kline_from_server || kline_for_mask)
		num_commands++;
	if (unkline_from_oper || unkline_from_server || unkline_for_mask)
		num_commands++;

	if (!num_commands)
	{
		syslog(LOG_NOTICE, "%s:%i section has nothing to allow",
				filename, linestart);
		return 0;
	}

	if (num_commands > 1)
	{
		syslog(LOG_NOTICE, "%s:%i: only 1 thing to allow can be specified",
			filename, linestart);
		return 0;
	}

	if (hostname || ipstring)
	{
		access_add(&(conf->allow), username, hostname, ipstring,
				class, password, -1, flags, 0, NULL, comment);
	}
	else if (gline_from_oper || gline_from_server || gline_for_mask)
	{
		IRCD_DENY_ACL	*acl;
		int				ret;

		if (!gline_from_oper)
			gline_from_oper = "*@*";
		if (!gline_from_server)
			gline_from_server = "*";
		if (!gline_for_mask)
			gline_for_mask = "*@*";

		ret = _deny_acl_add(&(_ircd_conf.gline_acls),
							&(_ircd_conf.gline_acls_last),
							gline_from_oper,
							gline_from_server,
							gline_for_mask,
							IRCD_DENY_ACL_FLAGS_ALLOW,
							&acl);
		if (ret)
		{
			syslog(LOG_NOTICE, "%s:%i gline_from_oper %s, gline_from_server %s, gline_for_mask %s already exists",
				filename, linestart, gline_from_oper, gline_from_server, gline_for_mask);
		}

		return 0;
	}
	else if (kline_from_oper || kline_from_server || kline_for_mask)
	{
		IRCD_DENY_ACL	*acl;
		int				ret;

		if (!kline_from_oper)
			kline_from_oper = "*@*";
		if (!kline_from_server)
			kline_from_server = "*";
		if (!kline_for_mask)
			kline_for_mask = "*@*";

		ret = _deny_acl_add(&(_ircd_conf.kline_acls),
							&(_ircd_conf.kline_acls_last),
							kline_from_oper,
							kline_from_server,
							kline_for_mask,
							IRCD_DENY_ACL_FLAGS_ALLOW,
							&acl);
		if (ret)
		{
			syslog(LOG_NOTICE, "%s:%i kline_from_oper %s, kline_from_server %s, kline_for_mask %s already exists",
				filename, linestart, kline_from_oper, kline_from_server, kline_for_mask);
		}

		return 0;
	}
	else if (unkline_from_oper || unkline_from_server || unkline_for_mask)
	{
		IRCD_DENY_ACL	*acl;
		int				ret;

		if (!unkline_from_oper)
			unkline_from_oper = "*@*";
		if (!unkline_from_server)
			unkline_from_server = "*";
		if (!unkline_for_mask)
			unkline_for_mask = "*@*";

		ret = _deny_acl_add(&(_ircd_conf.unkline_acls),
							&(_ircd_conf.unkline_acls_last),
							unkline_from_oper,
							unkline_from_server,
							unkline_for_mask,
							IRCD_DENY_ACL_FLAGS_ALLOW, &acl);
		if (ret)
		{
			syslog(LOG_NOTICE, "%s:%i unkline_from_oper %s, unkline_from_server %s, unkline_for_mask %s already exists",
				filename, linestart, unkline_from_oper, unkline_from_server, unkline_for_mask);
		}

		return 0;
	}

	return 0;
}

static int _conf_class(IRCDCONF *conf, CONFIGENTRY *ce, char *cmd)
{
	IRCD_CLASSENTRY		class;
	IRCD_CLASSENTRY		*ceptr;
	CONFIGENTRY			*cesave = ce;
	int					ret;
	CONFIGENTRY			*ce2;

	class.maxsendq = IRCDCONF_DEFAULT_SENDQ;
	class.num_on = 0;
	class.pingfreq = IRCDCONF_DEFAULT_PINGFREQ;
	class.connfreq = IRCDCONF_DEFAULT_CONNFREQ;
	class.maxclients = 0;
	class.flags = 0;
	class.floodnum = IRCDCONF_DEFAULT_FLOODNUM;
	class.floodtime = IRCDCONF_DEFAULT_FLOODTIME;
	class.floodexit = IRCDCONF_DEFAULT_FLOODEXIT;
	class.uhlimit = 0;
	class.hlimit = 0;
	class.uvlimit = 0;
	class.vlimit = 0;
	class.ipaddr_rate_time = 0;
	class.ipaddr_rate_num = 0;
	class.clients = NULL;
	class.maxchannels = -1;
	class.lastconnect = -1;
	class.send_buffer_size = 0;
	class.recv_buffer_size = 0;
	class.min_idle = -1;
	class.max_idle = -1;
	strcpy(class.hostname, "user0.on.");

	if (!ce->ce_vardata || !*ce->ce_vardata)
	{
		syslog(LOG_NOTICE, "%s:%i: no class specified",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		return -1;
	}
	else if (!ce->ce_entries)
	{
		syslog(LOG_NOTICE, "%s:%i: no section for class",
				ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		return -1;
	}

	{
		int sl = strlen(ce->ce_vardata);
		int m = sl < MAXCLASSNAME ? sl : MAXCLASSNAME;

		class.classname = (char *)MyMalloc(m + 1);
		support_strncpy(class.classname, ce->ce_vardata, m+1);
	}
	
	for(ce=ce->ce_entries;ce;ce=ce->ce_next)
	{
		if (!strcasecmp(ce->ce_varname, "inherit"))
		{
			char	*clname;

			if (!ce->ce_vardata || !*ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no class name specified",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}

			ret = class_find(&(_ircd_conf.classes), ce->ce_vardata, &ceptr);
			if (ret < 0)
			{
				syslog(LOG_NOTICE,
					"%s:%i: couldn't find class '%s' for inherit",
					ce->ce_fileptr->cf_filename,
					ce->ce_varlinenum,
					ce->ce_vardata);
				continue;
			}

			clname = class.classname;
			memcpy(&class, ceptr, sizeof(IRCD_CLASSENTRY));
			class.classname = clname;
			class.num_on = 0;
			class.clients = NULL;
		}
		else if (!strcasecmp(ce->ce_varname, "send_buffer_size") ||
					!strcasecmp(ce->ce_varname, "sendbuf"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no buffer size specified",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (ce->ce_vardatanum < 0)
				syslog(LOG_NOTICE,
					"%s:%i: ignoring invalid buffer size",
					ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
			else
				class.send_buffer_size = ce->ce_vardatanum;
		}
		else if (!strcasecmp(ce->ce_varname, "recv_buffer_size") ||
					!strcasecmp(ce->ce_varname, "recvbuf"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no buffer size specified",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (ce->ce_vardatanum < 0)
				syslog(LOG_NOTICE,
					"%s:%i: ignoring invalid buffer size",
					ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
			else
				class.recv_buffer_size = ce->ce_vardatanum;
		}
		else if (!strcasecmp(ce->ce_varname, "maxsendq"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no maxsendq specified",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (ce->ce_vardatanum < 0)
				syslog(LOG_NOTICE,
					"%s:%i: ignoring invalid maxsendq",
					ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
			else
				class.maxsendq = ce->ce_vardatanum;
		}
		else if (!strcasecmp(ce->ce_varname, "maxclients"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no maxclients specified",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (ce->ce_vardatanum < 0)
				syslog(LOG_NOTICE,
					"%s:%i: ignoring invalid maxclients",
					ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
			else
				class.maxclients = ce->ce_vardatanum;
		}
		else if (!strcasecmp(ce->ce_varname, "maxchannels"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no maxchannels specified",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if ((ce->ce_vardatanum < 0) || (ce->ce_vardatanum > MAXCHANNELSPERUSER))
				syslog(LOG_NOTICE,
					"%s:%i: ignoring invalid maxchannels",
					ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
			else
				class.maxchannels = ce->ce_vardatanum;
		}
		else if (!strcasecmp(ce->ce_varname, "hlimit"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no hlimit specified",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (ce->ce_vardatanum < 0)
				syslog(LOG_NOTICE,
					"%s:%i: ignoring invalid hlimit",
					ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
			else
				class.hlimit = ce->ce_vardatanum;
		}
		else if (!strcasecmp(ce->ce_varname, "uhlimit"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no uhlimit specified",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (ce->ce_vardatanum < 0)
				syslog(LOG_NOTICE,
					"%s:%i: ignoring invalid uhlimit",
					ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
			else
				class.uhlimit = ce->ce_vardatanum;
		}
		else if (!strcasecmp(ce->ce_varname, "vlimit"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no vlimit specified",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (ce->ce_vardatanum < 0)
				syslog(LOG_NOTICE,
					"%s:%i: ignoring invalid vlimit",
					ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
			else
				class.vlimit = ce->ce_vardatanum;
		}
		else if (!strcasecmp(ce->ce_varname, "uvlimit"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no uvlimit specified",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (ce->ce_vardatanum < 0)
				syslog(LOG_NOTICE,
					"%s:%i: ignoring invalid uvlimit",
					ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
			else
				class.uvlimit = ce->ce_vardatanum;
		}
		else if (!strcasecmp(ce->ce_varname, "connfreq"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no connfreq specified",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (ce->ce_vardatanum < 0)
				syslog(LOG_NOTICE,
					"%s:%i: ignoring invalid connfreq",
					ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
			else
				class.connfreq = ce->ce_vardatanum;
		}
		else if (!strcasecmp(ce->ce_varname, "pingfreq"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no pingfreq specified",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (ce->ce_vardatanum < 60)
				syslog(LOG_NOTICE,
					"%s:%i: ignoring invalid pingfreq",
					ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
			else
				class.pingfreq = ce->ce_vardatanum;
		}
		else if (!strcasecmp(ce->ce_varname, "hostname"))
		{
			if (ce->ce_vardatanum)
				sprintf(class.hostname, "user%i.on.", ce->ce_vardatanum);
			else if (ce->ce_vardata && (strchr(ce->ce_vardata, '.') != NULL))
			{
				if (hostname_check(ce->ce_vardata))
				{
					support_strncpy(class.hostname, ce->ce_vardata, sizeof(class.hostname));
				}
				else
					strcpy(class.hostname, "user0.on.");
			}
			else
				strcpy(class.hostname, "user0.on.");
			class.flags |= CLASS_FLAGS_HIDEHOSTNAME;
		}
		else if (!strcasecmp(ce->ce_varname, "pingfreq"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no pingfreq specified",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (ce->ce_vardatanum < 60)
				syslog(LOG_NOTICE,
					"%s:%i: ignoring invalid pingfreq",
					ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
			else
				class.pingfreq = ce->ce_vardatanum;
		}
		else if (!strcasecmp(ce->ce_varname, "min_idle"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no minimum idle time specified",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (ce->ce_vardatanum < 0)
			{
				syslog(LOG_NOTICE,
					"%s:%i: ignoring invalid minimum idle time",
					ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
			}
			else if ((class.max_idle >= 0) &&
							(ce->ce_vardatanum > class.max_idle))
			{
				syslog(LOG_NOTICE,
					"%s:%i: ignoring -- minimum idle time is greater than maximum idle time",
					ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
			}
			else
			{
				class.min_idle = ce->ce_vardatanum;
			}
		}
		else if (!strcasecmp(ce->ce_varname, "max_idle"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no maximum idle time specified",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (ce->ce_vardatanum < 0)
			{
				syslog(LOG_NOTICE,
					"%s:%i: ignoring invalid maximum idle time",
					ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
			}
			else if ((class.min_idle >= 0) &&
							(ce->ce_vardatanum < class.min_idle))
			{
				syslog(LOG_NOTICE,
					"%s:%i: ignoring -- maximum idle time is less than minimum idle time",
					ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
			}
			else
			{
				class.max_idle = ce->ce_vardatanum;
			}
		}
		else if (!strcasecmp(ce->ce_varname, "flag"))
		{
			ce2 = ce->ce_entries ? ce->ce_entries : ce;
			for(;ce2;ce2=(ce2 == ce) ? NULL : ce2->ce_next)
			{
				char *flagname = (ce2 == ce) ? ce2->ce_vardata : ce2->ce_varname;
				if (!flagname)
				{
					syslog(LOG_NOTICE, "%s:%i: no flag given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
					continue;
				}
				if (!strcasecmp(flagname, "bypasslimits") ||
						!strcasecmp(flagname, "bypass_limits"))
					class.flags |= CLASS_FLAGS_BYPASSLIMITS;
				else if (!strcasecmp(flagname, "bypassbans") ||
								!strcasecmp(flagname, "bypass_deny"))
					class.flags |= CLASS_FLAGS_BYPASSDENY;
				else if (!strcasecmp(flagname, "identonly") ||
							!strcasecmp(flagname, "ident_only"))
					class.flags |= CLASS_FLAGS_IDENTONLY;
				else if (!strcasecmp(flagname, "notilde") ||
							!strcasecmp(flagname, "no_tilde"))
					class.flags |= CLASS_FLAGS_NOTILDE;
				else if (!strcasecmp(flagname, "hidehostname") ||
							!strcasecmp(flagname, "hide_hostname"))
					class.flags |= CLASS_FLAGS_HIDEHOSTNAME;
				else if (!strcasecmp(flagname, "bypassnicknamedeny") ||
						!strcasecmp(flagname, "bypass_nickname_deny"))
					class.flags |= CLASS_FLAGS_BYPASSNICKNAMEDENY;
				else if (!strcasecmp(flagname, "bypassircnamedeny") ||
						!strcasecmp(flagname, "bypass_ircname_deny"))
					class.flags |= CLASS_FLAGS_BYPASSIRCNAMEDENY;
				else if (!strcasecmp(flagname, "bypasschanneldeny") ||
						!strcasecmp(flagname, "bypass_channel_deny"))
					class.flags |= CLASS_FLAGS_BYPASSCHANNELDENY;
				else if (!strcasecmp(flagname, "random_idle"))
					class.flags |= CLASS_FLAGS_RANDOM_IDLE;
				else if (!strcasecmp(flagname, "hide_idle_from_opers"))
					class.flags |= CLASS_FLAGS_HIDE_IDLE_FROM_OPERS;
				else if (!strcasecmp(flagname, "whois_notice"))
					class.flags |= CLASS_FLAGS_WHOIS_NOTICE_LOCAL;
				else if (!strcasecmp(flagname, "whois_notice_global"))
					class.flags |= CLASS_FLAGS_WHOIS_NOTICE_GLOBAL;
				else
					syslog(LOG_NOTICE,
						"%s:%i: unknown flag",
						ce->ce_fileptr->cf_filename, ce2->ce_varlinenum);
			}
		}
		else if (!strcasecmp(ce->ce_varname, "floodtime"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no floodtime specified",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (ce->ce_vardatanum < 0)
				syslog(LOG_NOTICE,
					"%s:%i: ignoring invalid floodtime",
					ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
			else
				class.floodtime = ce->ce_vardatanum;
		}
		else if (!strcasecmp(ce->ce_varname, "floodnum"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no floodnum specified",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (ce->ce_vardatanum < 0)
				syslog(LOG_NOTICE,
					"%s:%i: ignoring invalid floodnum",
					ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
			else
				class.floodnum = ce->ce_vardatanum;
		}
		else if (!strcasecmp(ce->ce_varname, "floodexit"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no floodexit specified",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (ce->ce_vardatanum < 0)
				syslog(LOG_NOTICE,
					"%s:%i: ignoring invalid floodexit",
					ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
			else
				class.floodexit = ce->ce_vardatanum;
		}
		else if (!strcasecmp(ce->ce_varname, "ipaddr_rate_num"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no ipaddr_rate_num specified",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (ce->ce_vardatanum < 0)
				syslog(LOG_NOTICE,
					"%s:%i: ignoring invalid ipaddr_rate_num",
					ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
			else
				class.ipaddr_rate_num = ce->ce_vardatanum;
		}
		else if (!strcasecmp(ce->ce_varname, "ipaddr_rate_time"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no ipaddr_rate_time specified",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (ce->ce_vardatanum < 0)
				syslog(LOG_NOTICE,
					"%s:%i: ignoring invalid ipaddr_rate_time",
					ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
			else
				class.ipaddr_rate_time = ce->ce_vardatanum;
		}
		else
			syslog(LOG_NOTICE, "%s:%i: unknown option for class",
				ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
	}

	if ((class.max_idle == -1) &&
				(class.flags & CLASS_FLAGS_RANDOM_IDLE))
	{
		class.max_idle = Now;
		syslog(LOG_NOTICE, "%s:%i: using value of %d for missing max_idle option",
			cesave->ce_fileptr->cf_filename, cesave->ce_varlinenum,
			class.max_idle);
	}

	ret = class_add(&(conf->classes), &class);
	if (ret)
	{
		MyFree(class.classname);
	}
	return 0;
}

static int _conf_deny(IRCDCONF *conf, CONFIGENTRY *ce, char *cmd)
{
	char			*username = NULL;

	char			*hostname = NULL;
	char			*ipstring = NULL;

	char			*ircname = NULL;
	char			*nickname = NULL;
	char			*channel = NULL;
	char			*gline_for_mask = NULL;
	char			*gline_from_oper = NULL;
	char			*gline_from_server = NULL;
	char			*kline_for_mask = NULL;
	char			*kline_from_oper = NULL;
	char			*kline_from_server = NULL;
	char			*unkline_for_mask = NULL;
	char			*unkline_from_oper = NULL;
	char			*unkline_from_server = NULL;

	char			*reason = NULL;
	char			*comment = NULL;

	u_int			flags = 0;
	int				expires = -1;
	int				ret;
	CONFIGENTRY		*ce2;
	char			*filename = ce->ce_fileptr->cf_filename;
	int				linestart = ce->ce_sectlinenum;
	int				num_commands = 0;

	if (!ce->ce_entries)
	{
		syslog(LOG_NOTICE, "%s:%i: no section for deny",
				ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		return -1;
	}

	for(ce=ce->ce_entries;ce;ce=ce->ce_next)
	{
		if (!ce->ce_varname)
			continue;
		if (!strcasecmp(ce->ce_varname, "hostname"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no hostname given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!hostname)
				hostname = ce->ce_vardata;
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra hostname",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "ircname"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no ircname given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!ircname)
				ircname = ce->ce_vardata;
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra ircname",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "nickname"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no nickname given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!nickname)
				nickname = ce->ce_vardata;
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra nickname",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "channel"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no channel given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			else if (!IsChannel(ce->ce_vardata))
			{
				syslog(LOG_NOTICE,
						"%s:%i: invalid channel given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!channel)
				channel = ce->ce_vardata;
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra channel",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "username"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no username given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!username)
				username = ce->ce_vardata;
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra username",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "ipaddr"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no ipaddress given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!ipstring)
				ipstring = ce->ce_vardata;
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra ipaddress",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "reason"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no reason given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!reason)
			{
				reason = MyStrdup(ce->ce_vardata);
				if (strlen(reason) > 200)
					reason[200] = '\0';
			}
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra reason",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "flag"))
		{
			ce2 = ce->ce_entries ? ce->ce_entries : ce;
			for(;ce2;ce2=(ce2 == ce) ? NULL : ce2->ce_next)
			{
				char *flagname = (ce2 == ce) ? ce2->ce_vardata : ce2->ce_varname;
				if (!flagname)
				{
					syslog(LOG_NOTICE, "%s:%i: no flag given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
					continue;
				}
				syslog(LOG_NOTICE,
						"%s:%i: unknown flag",
						ce->ce_fileptr->cf_filename, ce2->ce_varlinenum);
			}
		}
		else if (!strcasecmp(ce->ce_varname, "expire"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no expire specified",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (expires == -1)
			{
				if (ce->ce_vardatanum <= 0)
					syslog(LOG_NOTICE,
						"%s:%i: ignoring invalid expire",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				else
					expires = ce->ce_vardatanum;
			}
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra expire",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "comment"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no comment given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!comment)
				comment = ce->ce_vardata;
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra comment",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "gline_for_mask"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no user@host given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!gline_for_mask)
				gline_for_mask = ce->ce_vardata;
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra 'gline_for_mask'",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "gline_from_oper"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no user@host given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!gline_from_oper)
				gline_from_oper = ce->ce_vardata;
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra 'gline_from_oper'",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "gline_from_server"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no server given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!gline_from_server)
				gline_from_server = ce->ce_vardata;
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra 'gline_from_server'",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "unkline_for_mask"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no user@host given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!unkline_for_mask)
				unkline_for_mask = ce->ce_vardata;
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra 'unkline_for_mask'",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "unkline_from_oper"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no user@host given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!unkline_from_oper)
				unkline_from_oper = ce->ce_vardata;
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra 'unkline_from_oper'",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "unkline_from_server"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no server given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!unkline_from_server)
				unkline_from_server = ce->ce_vardata;
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra 'unkline_from_server'",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else
			syslog(LOG_NOTICE, "%s:%i: unknown option for deny",
				ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
	}

	if (hostname || ipstring)
		num_commands++;
	if (ircname)
		num_commands++;
	if (nickname)
		num_commands++;
	if (channel)
		num_commands++;
	if (gline_from_oper || gline_from_server || gline_for_mask)
		num_commands++;
	if (kline_from_oper || kline_from_server || kline_for_mask)
		num_commands++;
	if (unkline_from_oper || unkline_from_server || unkline_for_mask)
		num_commands++;

	if (!num_commands)
	{
		syslog(LOG_NOTICE, "%s:%i section has nothing to deny",
				filename, linestart);
		if (reason)
			MyFree(reason);
		return 0;
	}

	if (num_commands > 1)
	{
		syslog(LOG_NOTICE, "%s:%i: only 1 thing to deny can be specified",
				filename, linestart);
		if (reason)
			MyFree(reason);
		return 0;
	}

	if (hostname || ipstring)
	{
		if ((expires == -1) || (expires > Now))
			access_add(&(conf->deny), username, hostname, ipstring,
				NULL, reason, expires, flags, 0, NULL, comment);
		if (reason)
			MyFree(reason);
		return 0;
	}
	else if (gline_from_oper || gline_from_server || gline_for_mask)
	{
		IRCD_DENY_ACL	*acl;

		if (!gline_from_oper)
			gline_from_oper = "*@*";
		if (!gline_from_server)
			gline_from_server = "*";
		if (!gline_for_mask)
			gline_for_mask = "*@*";

		ret = _deny_acl_add(&(_ircd_conf.gline_acls),
							&(_ircd_conf.gline_acls_last),
							gline_from_oper,
							gline_from_server,
							gline_for_mask,
							0, &acl);
		if (ret)
		{
			syslog(LOG_NOTICE, "%s:%i gline_from_oper %s, gline_from_server %s, gline_for_mask %s already exists",
				filename, linestart, gline_from_oper, gline_from_server, gline_for_mask);
		}

		if (reason)
			MyFree(reason);
		return 0;
	}
	else if (kline_from_oper || kline_from_server || kline_for_mask)
	{
		IRCD_DENY_ACL	*acl;

		if (!kline_from_oper)
			kline_from_oper = "*@*";
		if (!kline_from_server)
			kline_from_server = "*";
		if (!kline_for_mask)
			kline_for_mask = "*@*";

		ret = _deny_acl_add(&(_ircd_conf.kline_acls),
							&(_ircd_conf.kline_acls_last),
							kline_from_oper,
							kline_from_server,
							kline_for_mask,
							0, &acl);
		if (ret)
		{
			syslog(LOG_NOTICE, "%s:%i kline_from_oper %s, kline_from_server %s, kline_for_mask %s already exists",
				filename, linestart, kline_from_oper, kline_from_server, kline_for_mask);
		}

		if (reason)
			MyFree(reason);
		return 0;
	}
	else if (unkline_from_oper || unkline_from_server || unkline_for_mask)
	{
		IRCD_DENY_ACL	*acl;

		if (!unkline_from_oper)
			unkline_from_oper = "*@*";
		if (!unkline_from_server)
			unkline_from_server = "*";
		if (!unkline_for_mask)
			unkline_for_mask = "*@*";

		ret = _deny_acl_add(&(_ircd_conf.unkline_acls),
							&(_ircd_conf.unkline_acls_last),
							unkline_from_oper,
							unkline_from_server,
							unkline_for_mask,
							0, &acl);
		if (ret)
		{
			syslog(LOG_NOTICE, "%s:%i unkline_from_oper %s, unkline_from_server %s, unkline_for_mask %s already exists",
				filename, linestart, unkline_from_oper, unkline_from_server, unkline_for_mask);
		}

		if (reason)
			MyFree(reason);
		return 0;
	}
	else
	{
		STRINGDENY	**list;
		STRINGDENY	*sd;
		char		*name;

		if (ircname)
		{
			list = &(_ircd_conf.ircname_denies);
			name = ircname;
		}
		else if (nickname)
		{
			list = &(_ircd_conf.nickname_denies);
			name = nickname;
		}
		else if (channel)
		{
			list = &(_ircd_conf.channel_denies);
			name = channel;
		}
		else
		{
			list = NULL;
			name = NULL;	/* NOT REACHED */
		}

		sd = stringdeny_find(list, name, Now);
		if (sd)
		{
			syslog(LOG_NOTICE, "%s:%i stringdeny of %s already exists",
				filename, linestart, name);
			if (reason)
				MyFree(reason);
			return 0;
		}

		stringdeny_add(list, name, reason, expires, flags);
	}

	if (reason)
		MyFree(reason);

	return 0;
}

static int _conf_link(IRCDCONF *conf, CONFIGENTRY *ce, char *cmd)
{
	char		*hostname = NULL;
	char		*username = NULL;
	char		*server = NULL;
	char		*cpasswd = NULL;
	char		*npasswd = NULL;
	char		class[MAXCLASSNAME+1];
	int			flags = 0;
	int			port = 6667;
	int			hostmask = 0;
	ircd_uint32_t bindipaddr = -1;
	IRCD_STRINGLIST	*hubs = NULL;
	IRCD_STRINGLIST	*leafs = NULL;
	CONFIGENTRY	*ce2;
	CONFIGENTRY	*oldce = ce;
	int			linestart = ce->ce_sectlinenum;
	char		*filename = ce->ce_fileptr->cf_filename;

	if (!ce->ce_entries)
	{
		syslog(LOG_NOTICE, "%s:%i: no section of options",
				ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		return -1;
	}

	server = ce->ce_vardata;

	*class = '\0';

	for(ce=ce->ce_entries;ce;ce=ce->ce_next)
	{
		if (!ce->ce_varname)
			continue;
		if (!strcasecmp(ce->ce_varname, "name") ||
				!strcasecmp(ce->ce_varname, "server"))
		{
			if (!ce->ce_vardata || !*ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no server name specified",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!server)
			{
				server = ce->ce_vardata;
			}
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra server name",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "class"))
		{
			if (!ce->ce_vardata || !*ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no class specified",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!*class)
			{
				support_strncpy(class, ce->ce_vardata, sizeof(class));
			}
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra class",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "hostname") ||
					!strcasecmp(ce->ce_varname, "host"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no hostname given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!hostname)
				hostname = ce->ce_vardata;
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra hostname",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "bindip"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no ipaddress given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			bindipaddr = inet_addr(ce->ce_vardata);
			if (bindipaddr == -1)
			{
				syslog(LOG_NOTICE,
						"%s:%i: invalid ipaddress given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
		}
		else if (!strcasecmp(ce->ce_varname, "username"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no username given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!username)
				username = ce->ce_vardata;
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra username",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "cpasswd") ||
					!strcasecmp(ce->ce_varname, "send_password"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no cpasswd given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!cpasswd)
				cpasswd = ce->ce_vardata;
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra cpasswd",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "npasswd") ||
				!strcasecmp(ce->ce_varname, "accept_password"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no npasswd given",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!npasswd)
				npasswd = ce->ce_vardata;
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra npasswd",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "flag"))
		{
			ce2 = ce->ce_entries ? ce->ce_entries : ce;
			for(;ce2;ce2=(ce2 == ce) ? NULL : ce2->ce_next)
			{
				char *flagname = (ce2 == ce) ? ce2->ce_vardata : ce2->ce_varname;
				if (!flagname)
				{
					syslog(LOG_NOTICE, "%s:%i: no flag given",
						ce->ce_fileptr->cf_filename, ce2->ce_varlinenum);
					continue;
				}
				if (!strcasecmp(flagname, "autoconn"))
					flags |= SERVLINK_FLAGS_AUTOCONN;
				else if (!strcasecmp(flagname, "hidehostname"))
					flags |= SERVLINK_FLAGS_HIDEHOSTNAME;
				else
					syslog(LOG_NOTICE,
						"%s:%i: unknown flag",
						ce->ce_fileptr->cf_filename, ce2->ce_varlinenum);
			}
		}
		else if (!strcasecmp(ce->ce_varname, "autoconn"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: 'yes' or 'no' required",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!strcasecmp(ce->ce_varname, "yes") ||
					!strcasecmp(ce->ce_varname, "on"))
			{
				flags |= SERVLINK_FLAGS_AUTOCONN;
			}
			else if (!strcasecmp(ce->ce_varname, "no") ||
					!strcasecmp(ce->ce_varname, "off"))
			{
				flags |= ~SERVLINK_FLAGS_AUTOCONN;
			}
			else
			{
				syslog(LOG_NOTICE,
						"%s:%i: 'yes' or 'no' required",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
			}
		}
		else if (!strcasecmp(ce->ce_varname, "port"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no port specified",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (ce->ce_vardatanum <= 0)
				syslog(LOG_NOTICE,
					"%s:%i: ignoring invalid port",
					ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
			else
				port = ce->ce_vardatanum;
		}
		else if (!strcasecmp(ce->ce_varname, "hostmask"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no hostmask specified",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!hostmask)
			{
				if (ce->ce_vardatanum < 0)
					syslog(LOG_NOTICE,
						"%s:%i: ignoring invalid hostmask",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				else
					hostmask = ce->ce_vardatanum;
			}
			else
				syslog(LOG_NOTICE,
						"%s:%i: ignoring extra hostmask",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		}
		else if (!strcasecmp(ce->ce_varname, "hub") ||
					!strcasecmp(ce->ce_varname, "hub_mask"))
		{
			ce2 = ce->ce_entries ? ce->ce_entries : ce;
			for(;ce2;ce2=(ce2 == ce) ? NULL : ce2->ce_next)
			{
				char *hub = (ce2 == ce) ? ce2->ce_vardata : ce2->ce_varname;

				if (!hub)
				{
					syslog(LOG_NOTICE, "%s:%i: no hub mask given",
						ce->ce_fileptr->cf_filename, ce2->ce_varlinenum);
					continue;
				}
				stringlist_add(&hubs, hub);
			}
		}
		else if (!strcasecmp(ce->ce_varname, "leaf") ||
				!strcasecmp(ce->ce_varname, "leaf_mask"))
		{
			ce2 = ce->ce_entries ? ce->ce_entries : ce;
			for(;ce2;ce2=(ce2 == ce) ? NULL : ce2->ce_next)
			{
				char *leaf = (ce2 == ce) ? ce2->ce_vardata : ce2->ce_varname;

				if (!leaf)
				{
					syslog(LOG_NOTICE, "%s:%i: no leaf mask given",
						ce->ce_fileptr->cf_filename, ce2->ce_varlinenum);
					continue;
				}
				stringlist_add(&leafs, leaf);
			}
		}
		else
			syslog(LOG_NOTICE, "%s:%i: unknown option for link",
				ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
	}

	if (!server)
	{
		syslog(LOG_NOTICE,
			"%s:%i: no server name specified",
			oldce->ce_fileptr->cf_filename, oldce->ce_varlinenum);
	}

	if (!hostname)
	{
		syslog(LOG_NOTICE, "%s:%i: section has no hostname",
				filename, linestart);
		stringlist_free(&hubs);
		stringlist_free(&leafs);

		return 0;
	}

	if (servlink_add(&(conf->servers), server, hostname,
				username, cpasswd, npasswd, hostmask,
				port, flags, 0, hubs, leafs, class, bindipaddr))
	{
		stringlist_free(&hubs);
		stringlist_free(&leafs);
	}

	return 0;
}

static int _conf_listen(IRCDCONF *conf, CONFIGENTRY *ce, char *cmd)
{
	int				*ports = NULL;
	ircd_uint32_t	ipaddr = -1;
	int				numports = 0;
	int				tport;
	int				flags = 0;
	CONFIGENTRY		*ce2;
	CONFIGENTRY		*ce3;

	if (!ce->ce_entries)
	{
		syslog(LOG_NOTICE, "%s:%i: no section for listen",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		return -1;
	}
	for(ce=ce->ce_entries;ce;ce=ce->ce_next)
	{
		if (!strcasecmp(ce->ce_varname, "address"))
		{
			if (!ce->ce_entries)
			{
				syslog(LOG_NOTICE,
						"%s:%i: no section for address specified",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!ce->ce_vardata || (*ce->ce_vardata == '*'))
				ipaddr = INADDR_ANY;
			else
			{
				ipaddr = inet_addr(ce->ce_vardata);

				if (ipaddr == -1)
				{
					syslog(LOG_NOTICE, "%s:%i: invalid ipaddress. ignoring section",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
					continue;
				}
				ipaddr = ntohl(ipaddr);
			}
			for(ce2=ce->ce_entries;ce2;ce2=ce2->ce_next)
			{
				if (!ce2->ce_varname)
					continue;
				if (!strcasecmp(ce2->ce_varname, "port"))
				{
					if (!ce2->ce_vardata && !ce2->ce_entries)
					{
						syslog(LOG_NOTICE,
							"%s:%i: no port entries",
							ce2->ce_fileptr->cf_filename, ce2->ce_varlinenum);
						continue;
					}
					if (ce2->ce_vardata)
					{
						if (ce2->ce_vardatanum <= 0)
						{
							syslog(LOG_NOTICE,
								"%s:%i: ignoring invalid port",
								ce2->ce_fileptr->cf_filename, ce2->ce_varlinenum);
							continue;
						}
						else
						{
							if (!(numports % 5))
							{
								ports = (int *)MyRealloc(ports, sizeof(int)*(numports+6));
							}
							ports[numports] = ce2->ce_vardatanum;
							numports++;
						}
					}
					for(ce3=ce2->ce_entries;ce3;ce3=ce3->ce_next)
					{
						if (!ce3->ce_varname)
							continue;
						tport = atoi(ce3->ce_varname);
						if (tport <= 0)
						{
							syslog(LOG_NOTICE,
								"%s:%i: ignoring invalid port",
								ce3->ce_fileptr->cf_filename, ce3->ce_varlinenum);
							continue;
						}
						else
						{
							if (!(numports % 5))
							{
								ports = (int *)MyRealloc(ports, sizeof(int)*(numports+6));
							}
							ports[numports] = tport;
							numports++;
						}
					}
				}
				else
					syslog(LOG_NOTICE, "%s:%i: unknown option",
						ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
			}
		}
		else
			syslog(LOG_NOTICE, "%s:%i: unknown option",
				ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		if (numports != 0)
		{
			numports--;
			while(numports >= 0)
			{
				listener_add(&(conf->listeners), ports[numports],
							ipaddr, flags, 0);
				numports--;
			}
			MyFree(ports);
			ports = NULL;
			numports = 0;
		}
	}
	return 0;
}

static int _conf_global(IRCDCONF *conf, CONFIGENTRY *ce, char *cmd)
{
	if (!ce->ce_entries)
	{
		syslog(LOG_NOTICE, "%s:%i: no section for global",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		return -1;
	}

	for(ce=ce->ce_entries;ce;ce=ce->ce_next)
	{
		if (!ce->ce_varname)
			continue;
		if (!strcasecmp(ce->ce_varname, "name"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE, "%s:%i No name given",
					ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (!conf->name)
				conf->name = MyStrdup(ce->ce_vardata);
		}
		else if (!strcasecmp(ce->ce_varname, "info"))
		{
			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE, "%s:%i No info given",
					ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}
			if (conf->info)
				MyFree(conf->info);
			conf->info = MyStrdup(ce->ce_vardata);
		}
		else if (!strcasecmp(ce->ce_varname, "nameserver"))
		{
			void	*tptr;
			char	*ptr;
			int		i;
			u_int	ipaddr;
			u_short port = 53;

			if (!ce->ce_vardata)
			{
				syslog(LOG_NOTICE, "%s:%i No nameserver given",
					ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
				continue;
			}

			ptr = strchr(ce->ce_vardata, ':');
			if (ptr)
				*ptr++ = '\0';
			ipaddr = inet_addr(ce->ce_vardata); 
			if (ipaddr == -1)
			{
				if (ptr)
					*--ptr = ':';
				syslog(LOG_NOTICE, "%s:%i: invalid name server: %s",
					ce->ce_fileptr->cf_filename, ce->ce_varlinenum,
					ce->ce_varname);
				continue;
			}
			if (ptr)
				port = htons(atoi(ptr));
			else
				port = htons(port);
			tptr = MyRealloc(conf->ns_list, (conf->ns_num+1)*sizeof(struct sockaddr_in));
			if (!tptr)
				continue;
			conf->ns_list = (struct sockaddr_in *)tptr;
			for(i=0;i<conf->ns_num;i++)
				if ((conf->ns_list[i].sin_addr.s_addr == ipaddr) &&
						(conf->ns_list[i].sin_port == port))
					break;
			if (i != conf->ns_num)
			{
				if (ptr)
					*--ptr = ':';
				syslog(LOG_NOTICE, "%s:%i: duplicate name server: %s",
					ce->ce_fileptr->cf_filename, ce->ce_varlinenum,
					ce->ce_varname);
				continue;
			}
			conf->ns_list = (struct sockaddr_in *)tptr;
			conf->ns_list[conf->ns_num].sin_addr.s_addr = ipaddr;
			conf->ns_list[conf->ns_num++].sin_port = port;
		}
		else
			syslog(LOG_NOTICE, "%s:%i: unknown option",
				ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
	}
	if (!conf->info)
		conf->info = MyStrdup("No information specified");
	return 0;
}

static int _conf_opers(IRCDCONF *conf, CONFIGENTRY *ce, char *cmd)
{
	char		*hostname = NULL;
	char		*username = NULL;
	char		**hostnames = NULL;
	char		**usernames = NULL;
	char		*passwd = NULL;
	char		*name;
	char		class[MAXCLASSNAME+1];
	int			flags;
	int			userflags;
	int			numhosts = 0;
	CONFIGENTRY	*ce2;
	CONFIGENTRY	*ce3;
	int			i;

	if (!ce->ce_entries)
	{
		syslog(LOG_NOTICE, "%s:%i: no section for opers",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		return -1;
	}
	for(ce=ce->ce_entries;ce;ce=ce->ce_next)
	{
		if (!ce->ce_varname)
			continue;
		/* only option is 'name' so far. */
		if (strcasecmp(ce->ce_varname, "name"))
		{
			syslog(LOG_NOTICE, "%s:%i: unknown option",
				ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
			continue;
		}
		if (!ce->ce_entries)
		{
			syslog(LOG_NOTICE, "%s:%i: no section for oper name",
				ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
			continue;
		}
		else if (!ce->ce_vardata)
		{
			syslog(LOG_NOTICE, "%s:%i no name specified",
					ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
			continue;
		}
		name = ce->ce_vardata;
		flags = userflags = 0;
		*class = '\0';
		for(ce2=ce->ce_entries;ce2;ce2=ce2->ce_next)
		{
			if (!strcasecmp(ce2->ce_varname, "class"))
			{
				if (!ce2->ce_vardata || !*ce2->ce_vardata)
				{
					syslog(LOG_NOTICE,
							"%s:%i: no class specified",
							ce2->ce_fileptr->cf_filename, ce2->ce_varlinenum);
					continue;
				}
				if (!*class)
				{
					support_strncpy(class, ce2->ce_vardata, sizeof(class));
				}
				else
					syslog(LOG_NOTICE,
							"%s:%i: ignoring extra class",
							ce2->ce_fileptr->cf_filename, ce2->ce_varlinenum);
			}
			else if (!strcasecmp(ce2->ce_varname, "from"))
			{
				if (!ce2->ce_entries)
				{
					syslog(LOG_NOTICE,
							"%s:%i: no section for from",
							ce2->ce_fileptr->cf_filename, ce2->ce_varlinenum);
					continue;
				}
				if (!(numhosts % 5))
				{
					hostnames = (char **)MyRealloc(hostnames, sizeof(char *)*(numhosts+6));
					usernames = (char **)MyRealloc(usernames, sizeof(char *)*(numhosts+6));
					hostnames[numhosts] = usernames[numhosts] = NULL;
				}
				for(ce3=ce2->ce_entries;ce3;ce3=ce3->ce_next)
				{
					if (!ce3->ce_varname)
						continue;
					if (!strcasecmp(ce3->ce_varname, "hostname"))
					{
						if (!ce3->ce_vardata)
						{
							syslog(LOG_NOTICE,
									"%s:%i: no hostname given",
									ce3->ce_fileptr->cf_filename, ce3->ce_varlinenum);
							continue;
						}
						if (!hostnames[numhosts])
							hostnames[numhosts] = ce3->ce_vardata;
						else
							syslog(LOG_NOTICE,
									"%s:%i: ignoring extra hostname",
									ce3->ce_fileptr->cf_filename, ce3->ce_varlinenum);
					}
					else if (!strcasecmp(ce3->ce_varname, "username"))
					{
						if (!ce3->ce_vardata)
						{
							syslog(LOG_NOTICE,
									"%s:%i: no username given",
									ce3->ce_fileptr->cf_filename, ce3->ce_varlinenum);
							continue;
						}
						if (!usernames[numhosts])
							usernames[numhosts] = ce3->ce_vardata;
						else
							syslog(LOG_NOTICE,
									"%s:%i: ignoring extra username",
									ce3->ce_fileptr->cf_filename, ce3->ce_varlinenum);
					}
					else
						syslog(LOG_NOTICE,
							"%s:%i: unknown option for from",
							ce3->ce_fileptr->cf_filename, ce3->ce_varlinenum);
				}
				if (!hostnames[numhosts])
				{
					usernames[numhosts] = NULL;
					continue;
				}
				numhosts++;
				hostnames[numhosts] = usernames[numhosts] = NULL;
			}
			else if (!strcasecmp(ce2->ce_varname, "hostname"))
			{
				if (!ce2->ce_vardata)
				{
					syslog(LOG_NOTICE,
							"%s:%i: no hostname given",
							ce2->ce_fileptr->cf_filename, ce2->ce_varlinenum);
					continue;
				}
				if (!hostname)
					hostname = ce2->ce_vardata;
				else
					syslog(LOG_NOTICE,
							"%s:%i: ignoring extra hostname",
							ce2->ce_fileptr->cf_filename, ce2->ce_varlinenum);
			}
			else if (!strcasecmp(ce2->ce_varname, "username"))
			{
				if (!ce2->ce_vardata)
				{
					syslog(LOG_NOTICE,
							"%s:%i: no username given",
							ce2->ce_fileptr->cf_filename, ce2->ce_varlinenum);
					continue;
				}
				if (!username)
					username = ce2->ce_vardata;
				else
					syslog(LOG_NOTICE,
							"%s:%i: ignoring extra username",
							ce2->ce_fileptr->cf_filename, ce2->ce_varlinenum);
			}
			else if (!strcasecmp(ce2->ce_varname, "passwd"))
			{
				if (!ce2->ce_vardata)
				{
					syslog(LOG_NOTICE,
							"%s:%i: no passwd given",
							ce2->ce_fileptr->cf_filename, ce2->ce_varlinenum);
					continue;
				}
				if (!passwd)
					passwd = ce2->ce_vardata;
				else
					syslog(LOG_NOTICE,
							"%s:%i: ignoring extra passwd",
							ce2->ce_fileptr->cf_filename, ce2->ce_varlinenum);
			}
			else if (!strcasecmp(ce2->ce_varname, "flag"))
			{
				ce3 = ce2->ce_entries ? ce2->ce_entries : ce2;
				for(;ce3;ce3=(ce3 == ce2) ? NULL : ce3->ce_next)
				{
					char *flagname = (ce3 == ce2) ? ce3->ce_vardata : ce3->ce_varname;
					if (!flagname)
					{
						syslog(LOG_NOTICE, "%s:%i: no flag given",
							ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
						continue;
					}
					if (!strcasecmp(flagname, "local"))
						flags |= OLINE_FLAGS_LOCAL;
					else if (!strcasecmp(flagname, "cankline") ||
								!strcasecmp(flagname, "can_kline"))
						flags |= OLINE_FLAGS_CANKLINE;
					else if (!strcasecmp(flagname, "can_remote_kline"))
						flags |= OLINE_FLAGS_CANREMOTEKLINE;
					else if (!strcasecmp(flagname, "canunkline") ||
								!strcasecmp(flagname, "can_unkline"))
						flags |= OLINE_FLAGS_CANUNKLINE;
					else if (!strcasecmp(flagname, "can_remote_unkline"))
						flags |= OLINE_FLAGS_CANREMOTEUNKLINE;
					else if (!strcasecmp(flagname, "cangline") ||
								!strcasecmp(flagname, "can_gline"))
						flags |= OLINE_FLAGS_CANGLINE;
					else if (!strcasecmp(flagname, "canlocalroute") ||
								!strcasecmp(flagname, "can_local_route"))
						flags |= OLINE_FLAGS_CANLOCALROUTE;
					else if (!strcasecmp(flagname, "canglobalroute") ||
								!strcasecmp(flagname, "can_global_route"))
						flags |= OLINE_FLAGS_CANGLOBALROUTE;
					else if (!strcasecmp(flagname, "canlocalkill") ||
								!strcasecmp(flagname, "can_local_kill"))
						flags |= OLINE_FLAGS_CANLOCALKILL;
					else if (!strcasecmp(flagname, "canglobalkill") ||
								!strcasecmp(flagname, "can_global_kill"))
						flags |= OLINE_FLAGS_CANGLOBALKILL;
					else if (!strcasecmp(flagname, "canflood") ||
								!strcasecmp(flagname, "can_flood"))
						flags |= OLINE_FLAGS_CANFLOOD;
					else if (!strcasecmp(flagname, "canset") ||
								!strcasecmp(flagname, "can_set"))
						flags |= OLINE_FLAGS_CANSET;
					else if (!strcasecmp(flagname, "candie") ||
								!strcasecmp(flagname, "can_die"))
						flags |= OLINE_FLAGS_CANDIE;
					else if (!strcasecmp(flagname, "canrehash") ||
								!strcasecmp(flagname, "can_rehash"))
						flags |= OLINE_FLAGS_CANREHASH;
					else if (!strcasecmp(flagname, "can_restart"))
						flags |= OLINE_FLAGS_CANRESTART;
					else
						syslog(LOG_NOTICE,
							"%s:%i: unknown flag",
							ce->ce_fileptr->cf_filename, ce3->ce_varlinenum);
				}
			}
			else if (!strcasecmp(ce2->ce_varname, "userflag"))
			{
				IRCD_USERFLAG	*u;

				ce3 = ce2->ce_entries ? ce2->ce_entries : ce2;
				for(;ce3;ce3=(ce3 == ce2) ? NULL : ce3->ce_next)
				{
					char *flagname = (ce3 == ce2) ? ce3->ce_vardata : ce3->ce_varname;
					if (!flagname)
					{
						syslog(LOG_NOTICE, "%s:%i: no flag given",
							ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
						continue;
					}

					for(u=_user_flags;u->flagstr;u++)
						if (!strcasecmp(u->flagstr, flagname))
							break;
					if (u)
						userflags |= u->flag;
					else
						syslog(LOG_NOTICE,
							"%s:%i: unknown user flag",
							ce->ce_fileptr->cf_filename, ce3->ce_varlinenum);
				}
			}
		}
		if ((hostname || hostnames) && name && passwd)
		{
			if (hostname)
			{
				oline_add(&(conf->olines), hostname,
					username, name, passwd, flags, 0, class, userflags);
			}
			if (hostnames)
			{
				for(i=0;hostnames[i];i++)
					oline_add(&(conf->olines), hostnames[i],
						usernames[i], name, passwd, flags, 0, class, userflags);
			}
		}

		hostname = NULL;
		username = NULL;
		passwd = NULL;

		if (hostnames)
		{
			MyFree(hostnames);
			MyFree(usernames);
			hostnames = usernames = NULL;
			numhosts = 0;
		}
	}
	return 0;
}

static int _conf_set(IRCDCONF *conf, CONFIGENTRY *ce, char *cmd)
{
	char		*args;
	IRCD_VAR	*vptr;

	if (!ce->ce_entries)
	{
		syslog(LOG_NOTICE, "%s:%i: no section for set",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		return -1;
	}
	for(ce=ce->ce_entries;ce;ce=ce->ce_next)
	{
		if (!ce->ce_varname)
			continue;
		vptr = vars_find_byname(ce->ce_varname);
		if (!vptr)
			continue;
		args = ce->ce_vardata;
		if (vptr->v_type == IRCD_VAR_TYPE_INT)
			args = (char *)args ? (char *)ce->ce_vardatanum : (char *)0;
		else if (vptr->v_type == IRCD_VAR_TYPE_BOOL)
		{
			if (!args || !*args || !strncasecmp(args, "yes", 3) || !strncasecmp(args, "on", 2))
				args = (char *)1;
			else if (!strncasecmp(args, "no", 2) || !strncasecmp(args, "off", 3))
				args = (char *)0;
			else
				args = atoi(args) ? (char *)1 : (char *)0;
		}
		vars_value_set_byname(NULL, ce->ce_varname, args);
    }
	return 0;
}

int ircdconf_prepare_rehash(IRCDCONF *ircdconf)
{
	listener_clear(&(ircdconf->listeners));
	class_clear(&(ircdconf->classes));
	servlink_clear(&(ircdconf->servers));
	oline_clear(&(ircdconf->olines));
	access_clear(&(ircdconf->allow), 0);
	access_clear(&(ircdconf->deny), 0);
	stringdeny_free(&(ircdconf->nickname_denies));
	stringdeny_free(&(ircdconf->ircname_denies));
	stringdeny_free(&(ircdconf->channel_denies));
	_deny_acls_free(&(ircdconf->kline_acls), &(ircdconf->kline_acls_last));
	_deny_acls_free(&(ircdconf->unkline_acls), &(ircdconf->unkline_acls_last));
	motd_unload(&Motd);
	logs_flushall();
	return 0;
}

int ircdconf_rehash(void)
{
	int	 			ret;
	struct sockaddr_in	*sin;
	int					sin_num;
	

	ircdconf_prepare_rehash(&_ircd_conf);

	sin = _ircd_conf.ns_list;
	sin_num = _ircd_conf.ns_num;

	_ircd_conf.ns_list = NULL;
	_ircd_conf.ns_num = 0;

	ret = ircdconf_read(Conffile, &_ircd_conf);
	if (ret)
	{
		send_to_opers("Error reading config file: %i\n", ret);
		_ircd_conf.ns_list = sin;
		_ircd_conf.ns_num = sin_num;
	}
	else
	{
		char	*motd_file = vars_value_get_bynum(IRCD_VAR_MOTD_FILE);

		if (_ircd_conf.info && *_ircd_conf.info)
			support_strncpy(meptr->ircname, _ircd_conf.info,
					sizeof(meptr->ircname));
		if (motd_file)
			motd_load(motd_file, &Motd);
		logs_init();
		if (!_ircd_conf.ns_num)
		{
			_ircd_conf.ns_list = sin;
			_ircd_conf.ns_num = sin_num;
		}
		else if (sin)
			MyFree(sin);
		build_isupport();
	}

	if (RehashWho)
	{
		if (!ret)
			send_to_opers("Server config file has been rehashed by %s",
				RehashWho);
		MyFree(RehashWho);
		RehashWho = NULL;
	}

	Rehashed = 1;

	return ret;
}

int ircdconf_init(IRCDCONF *ircdconf)
{
	BZERO(ircdconf, sizeof(IRCDCONF));
	access_init(&(ircdconf->allow));
	access_init(&(ircdconf->deny));
	return 0;
}

int ircdconf_read(char *filename, IRCDCONF *ircdconf)
{
	CONFIGFILE		*cfptr;
	CONFIGFILE		*cfp;
	CONFIGENTRY		*ce;
	CONFCOMMAND		*cf;
	SERVLINK		*sl;

	cfptr = cfp = config_load(filename, NULL);
	if (!cfp)
		return -ENOENT;
	for(;cfptr;cfptr=cfptr->cf_next)
	for(ce=cfptr->cf_entries;ce;ce=ce->ce_next)
	{
		if (!ce->ce_varname)
			continue;
		for(cf=_ConfCommands;cf->name;cf++)
		{
			if (!strcasecmp(cf->name, ce->ce_varname))
			{
				cf->func(ircdconf, ce, cf->name);
				break;
			}
		}
		if (!cf)
		{
			syslog(LOG_NOTICE, "%s:%i: invalid config option: %s",
				ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ce->ce_varname);
		}
	}
	config_free(cfp);
	if (ircdconf->ns_num)
	{
		ircdres_set_ns((IRCDRES *)(Ircd_resconn->conn_ptr), ircdconf->ns_list, ircdconf->ns_num);
	}
	else
	{
		syslog(LOG_ERR, "warning: no nameservers listed in conf files");
	}

	for(sl=ircdconf->servers;sl;sl=sl->next)
	{
		if ((sl->inaddr.s_addr == -1) && !(sl->flags & SERVLINK_FLAGS_DOINGDNS))
		{
			IRCDRES_HOSTENT	*heptr;
			int				ret;

			sl->res_req = (IRCDRES_REQLIST *)MyMalloc(sizeof(IRCDRES_REQLIST));
			ret = ircdres_gethostbyname((IRCDRES *)(Ircd_resconn->conn_ptr), sl->hostname,
						(ircdres_func_t)_servlink_res_callback,
						(void *)sl, &heptr, &sl->res_req->rl_req);
			if ((ret >= 0) && heptr)
			{
				if (heptr->he_he.h_length == sizeof(struct in_addr))
					BCOPY(heptr->he_he.h_addr_list[0], &sl->inaddr,
							sizeof(struct in_addr));
				MyFree(heptr);
				MyFree(sl->res_req);
				sl->res_req = NULL;
			}
			else if (ret == -EINPROGRESS)
			{
				sl->flags |= SERVLINK_FLAGS_DOINGDNS;
				sl->res_starttime = Now;
			}
			else
			{
				send_to_opers("Lookup of %s for \"link %s\" conf entry failed: %s",
					sl->hostname, sl->servername, support_strerror(-ret));
				MyFree(sl->res_req);
				sl->res_req = NULL;
			}
		}
	}
	return 0;
}

int listener_clear(IRCD_LISTENER **lptr)
{
	IRCD_LISTENER	*l;
	IRCD_LISTENER	*n;

	if (!lptr)
		return -EINVAL;

	for(l=*lptr;l;l=n)
	{
		n = l->l_next;
		l->l_flags |= IRCD_LISTENER_FLAGS_ILLEGAL;
		if (!l->l_num_on)
		{
			del_listener_from_llist(lptr, l);
			ircd_close_fd(&(l->l_connptr->fdlistinfo));
			MyFree(l->l_connptr);
			MyFree(l);
		}
	}
	return 0;
}

int listener_add(IRCD_LISTENER **lptr, int port, ircd_uint32_t ipaddr, int flags, int aflags)
{
	IRCD_LISTENER	*l;
	int				fd;
	int				ret;

	if (!lptr || (port == -1))
		return -EINVAL;

	for(l=*lptr;l;l=l->l_next)
	{
		if ((l->l_port == port) && (l->l_ipaddr == ipaddr))
		{
			l->l_flags &= ~IRCD_LISTENER_FLAGS_ILLEGAL;
			return 0;
		}
	}

	fd = mysocket_listenport(ipaddr, port, MYSOCKET_FLAGS_REUSEADDR|MYSOCKET_FLAGS_NONBLOCKING);
	if (fd < 0)
		return fd;

	l = (IRCD_LISTENER *)MyMalloc(sizeof(IRCD_LISTENER));
	l->l_lasttime = l->l_firsttime = Now;
	strncpy(l->l_ipstring, inet_ntoa(*((struct in_addr *)&ipaddr)), sizeof(l->l_ipstring));
	l->l_ipaddr = ipaddr;
	l->l_port = port;
	l->l_flags = flags;
	l->l_num_on = 0;
	l->l_connptr = (IRCD_CONN *)MyMalloc(sizeof(IRCD_CONN));
	l->l_connptr->conn_type = IRCD_CONN_TYPE_LISTENER;
	l->l_connptr->conn_ptr = (void *)l;

	if ((ret = ircd_open_fd(fd, FDLIST_FLAGS_READ, l->l_connptr, NULL)) < 0)
	{
		close(fd);
		MyFree(l->l_connptr);
		MyFree(l);
		return ret;
	}
	add_listener_to_llist(lptr, l);
	return 0;
}

int servlink_clear(SERVLINK **slptr)
{
	SERVLINK	*n;

	if (!slptr)
		return -EINVAL;
	for(;*slptr;(*slptr) = n)
	{
		n = (*slptr)->next;
		if ((*slptr)->res_req) /* doing lookup */
		{
			ircdres_request_del((*slptr)->res_req->rl_req, 0);
			del_reqlist_from_list(&Reqlist, (*slptr)->res_req);
			MyFree((*slptr)->res_req);
			(*slptr)->res_req = NULL;
		}
		if ((*slptr)->servername)
			MyFree((*slptr)->servername);
		if ((*slptr)->hostname)
			MyFree((*slptr)->hostname);
		if ((*slptr)->username)
			MyFree((*slptr)->username);
		if ((*slptr)->cpasswd)
			MyFree((*slptr)->cpasswd);
		if ((*slptr)->npasswd)
			MyFree((*slptr)->npasswd);
		if ((*slptr)->classname)
			MyFree((*slptr)->classname);
		stringlist_free(&((*slptr)->hubs));
		stringlist_free(&((*slptr)->leafs));
		MyFree((*slptr));
	}
	return 0;
}

int servlink_add(SERVLINK **slptr, char *server, char *hostname,
		char *username, char *cpasswd, char *npasswd,
		int hostmask, int port, int flags, int aflags,
		IRCD_STRINGLIST *hubs, IRCD_STRINGLIST *leafs, char *class,
		ircd_uint32_t bindipaddr)
{
	SERVLINK		*sl;

	if (!slptr || !server || !hostname)
		return -EINVAL;

	sl = (SERVLINK *)MyMalloc(sizeof(SERVLINK));
	sl->servername = MyStrdup(server);
	sl->hostname = MyStrdup(hostname);
	sl->bindipaddr = bindipaddr == -1 ? -1 : ntohl(bindipaddr);
	sl->username = username ? MyStrdup(username) : NULL;
	sl->inaddr.s_addr = inet_addr(hostname);
	if (sl->inaddr.s_addr != -1)
		sl->inaddr.s_addr = ntohl(sl->inaddr.s_addr);
	sl->cpasswd = cpasswd ? MyStrdup(cpasswd) : NULL;
	sl->npasswd = npasswd ? MyStrdup(npasswd) : NULL;
	sl->hostmask = hostmask;
	sl->hubs = hubs;
	sl->leafs = leafs;
	sl->classname = MyStrdup(class);
	sl->port = port;
	sl->flags = flags;
	sl->flags &= ~SERVLINK_FLAGS_DOINGDNS;
	sl->next = *slptr;
	*slptr = sl;
	return 0;
}

SERVLINK *servlink_find(char *servername, IRCD_CLIENT *pptr)
{
	SERVLINK		*sl;
#if 0
	ircd_uint32_t	ipaddr = htonl(pptr->c_rem_ipaddr);
	char			**addrs;
#endif

	for(sl=_ircd_conf.servers;sl;sl=sl->next)
	{
		if (!mycmp(servername, sl->servername))
		{
			/* check ips now */
			if (sl->flags & SERVLINK_FLAGS_DOINGDNS)
				continue;
			if (sl->inaddr.s_addr != -1)
			{
				if (pptr->c_rem_ipaddr == sl->inaddr.s_addr)
					break;
			}
			else
			{
#if 0
				for(addrs=sl->he_forward.h.h_addr_list;*addrs;addrs++)
					if (!BCMP(*addrs, (char *)&inaddr, sizeof(struct in_addr)))
						break;
				if (*addrs)
					break;
#else
				continue;
#endif
			}
		}
	}
	return sl;
}

int oline_clear(OLINE **coptr)
{
	OLINE	*n;

	if (!coptr)
		return -EINVAL;

	for(;*coptr;*coptr=n)
	{
		n = (*coptr)->next;
		if ((*coptr)->hostname)
			MyFree((*coptr)->hostname);
		if ((*coptr)->username)
			MyFree((*coptr)->username);
		if ((*coptr)->name)
			MyFree((*coptr)->name);
		if ((*coptr)->ipstring)
			MyFree((*coptr)->ipstring);
		if ((*coptr)->passwd)
			MyFree((*coptr)->passwd);
		if ((*coptr)->classname)
			MyFree((*coptr)->classname);
		MyFree((*coptr));
	}
	return 0;
}

int oline_add(OLINE **coptr, char *hostname, char *username,
		char *name, char *passwd, int flags, int aflags, char *class, int userflags)
{
	OLINE	*co;

	if (!coptr || !name || !hostname)
		return -EINVAL;

	co = (OLINE *)MyMalloc(sizeof(OLINE));
	co->hostname = MyStrdup(hostname);
	co->username = username ? MyStrdup(username) : NULL;
	co->name = MyStrdup(name);
	co->passwd = passwd ? MyStrdup(passwd) : NULL;
	co->flags = flags;
	co->userflags = userflags;
	co->classname = MyStrdup(class);
	co->next = *coptr;
	*coptr = co;
	return 0;
}

int oline_find(OLINE **coptr, char *name, char *username, char *hostname, ircd_uint32_t ipaddr, OLINE **corptr)
{
	OLINE			*cptr;
	ircd_uint32_t	net;
	ircd_uint32_t	mask;

	for(cptr=*coptr;cptr;cptr=cptr->next)
	{
		if (!mycmp(name, cptr->name) &&
				(!cptr->username || !match(cptr->username, username)) &&
				(!match(cptr->hostname, hostname) ||
				((1 == cidr_ip2netmask(cptr->hostname, &net, &mask) &&
				!cidr_match(net, mask, ipaddr)))))

		{
			if (corptr)
				*corptr = cptr;
			return 0;
		}
	}
	return -ENOENT;
}

int class_del(CLASS *cptr, IRCD_CLASSENTRY *ceptr)
{
	int		i;

	for(i=0;i<cptr->num;i++)
		if (cptr->entries[i] == ceptr)
			break;
	assert(i != cptr->num);
	if (ceptr->classname)
		MyFree(ceptr->classname);
	MyFree(ceptr);
	if (i != cptr->num-1)
		BCOPY(&(cptr->entries[i+1]), &(cptr->entries[i]),
			sizeof(IRCD_CLASSENTRY *)*(cptr->num - i - 1));
	BZERO(&(cptr->entries[cptr->num-1]), sizeof(IRCD_CLASSENTRY *));
	cptr->num--;
	return 0;
}

int class_clear(CLASS *cptr)
{
	int		i = 0;

	if (!cptr)
		return -EINVAL;
	if (!cptr->num)
		return 0;
	for(;;)
	{
		cptr->entries[i]->flags |= CLASS_FLAGS_ILLEGAL;
		if (!cptr->entries[i]->num_on)
		{
			MyFree(cptr->entries[i]);
			if (i != cptr->num-1)
				BCOPY(&(cptr->entries[i+1]), &(cptr->entries[i]),
					sizeof(IRCD_CLASSENTRY *)*(cptr->num - i - 1));
			BZERO(&(cptr->entries[cptr->num-1]), sizeof(IRCD_CLASSENTRY *));
			if (--cptr->num == i)
				break;
		}
		else
		{
			if (++i == cptr->num)
				break;
		}
	}
	return 0;
}

int class_add(CLASS *cptr, IRCD_CLASSENTRY *ceptr)
{
	int		lower = 0;
	int		upper;
	int		compare;
	int		ret = 0;

	if (!cptr)
	{
		return -EINVAL;
	}

	if (cptr->size && cptr->num)
	{
		upper = cptr->num-1;
		compare = (lower+upper)/2;

		while(upper != lower)
		{
			ret = strcasecmp(cptr->entries[compare]->classname, ceptr->classname);

			if (!ret)
				break;
			else if (ret < 0)
				upper = compare;
			else
				lower = (lower == compare) ? compare + 1 : compare;
			compare = (lower+upper)/2;
		}
	}
	else
		compare = 0;

	if (cptr->num)
		ret = strcasecmp(cptr->entries[compare]->classname, ceptr->classname);

	if (!cptr->num || ret)
	{
		if (!cptr->size || ((cptr->num+1) == cptr->size))
		{
			cptr->size += 5;
			cptr->entries = (IRCD_CLASSENTRY **)MyRealloc(cptr->entries,
							cptr->size*sizeof(IRCD_CLASSENTRY *));
			BZERO((cptr->entries+cptr->num), sizeof(IRCD_CLASSENTRY *)*5);
		}
		if (cptr->num && (ret > 0))
			compare++;
		if (cptr->num - compare > 0)
		{
			BCOPY(&(cptr->entries[compare]), &(cptr->entries[compare+1]),
				sizeof(IRCD_CLASSENTRY *)*(cptr->num - compare));
		}
		cptr->entries[compare] = (IRCD_CLASSENTRY *)MyMalloc(sizeof(IRCD_CLASSENTRY));
		cptr->entries[compare]->num_on = 0;
		cptr->entries[compare]->clients = NULL;
		cptr->num++;
	}
	ceptr->num_on = cptr->entries[compare]->num_on;
	ceptr->clients = cptr->entries[compare]->clients;
	BCOPY(ceptr, cptr->entries[compare], sizeof(IRCD_CLASSENTRY));
	cptr->entries[compare]->flags &= ~CLASS_FLAGS_ILLEGAL; /* just in case */
	return 0;
}

int class_find(CLASS *cptr, char *class, IRCD_CLASSENTRY **classptr)
{
	int		lower = 0;
	int		upper;
	int		compare;
	int		ret;

	if (!cptr)
		return -EINVAL;

	if (!cptr->size || !cptr->num)
		return -ENOENT;

	upper = cptr->num-1;
	compare = (lower+upper)/2;

	while(upper != lower)
	{
		ret = strcasecmp(cptr->entries[compare]->classname, class);

		if (!ret)
			break;
		else if (ret < 0)
			upper = compare;
		else
			lower = (lower == compare) ? compare + 1 : compare;
		compare = (lower+upper)/2;
	}

	ret = strcasecmp(cptr->entries[compare]->classname, class);

	if (ret || (cptr->entries[compare]->flags & CLASS_FLAGS_ILLEGAL))
		return -ENOENT;
	if (classptr)
		*classptr = cptr->entries[compare];
	return 0;
}

int ircd_checkallow(IRCD_CLIENT *pptr, int *err)
{
	int					ret;
	CONFHOSTIPENTRY		*entryptr;
	IRCD_CLASSENTRY			*classptr;

	ret = access_match_ipaddr(&(_ircd_conf.allow),
				pptr->username, pptr->c_ipstring,
				pptr->c_rem_ipaddr, ACCESS_FLAGS_BESTMATCH,
				&entryptr);
	if (strcmp(pptr->c_ipstring, pptr->hostname) &&
		((ret < 0) || (entryptr->flags & ALLOW_FLAGS_CHECKHOSTNAME)))
	{
		CONFHOSTIPENTRY		*eptr;
		int				r;

		r = access_match_hostname(&(_ircd_conf.allow),
				pptr->username, pptr->hostname, ACCESS_FLAGS_BESTMATCH,
				&eptr);
		if (!r)
		{
			entryptr = eptr;
			ret = r;
		}
	}

	/* FIX...check pptr->res_req->rl_reverse aliases also */

	if (ret < 0)
		ret = access_match_ipaddr_default(&(_ircd_conf.allow),
				pptr->username, pptr->c_ipstring,
				pptr->c_rem_ipaddr, ACCESS_FLAGS_BESTMATCH,
				&entryptr);
	if (ret < 0)
	{
		if (err)
			*err = ret;
		return -1;
	}

	if (entryptr->data &&
		(!pptr->c_passwd || strcmp(entryptr->data, pptr->c_passwd)))
	{
		return -8;
	}

	if (entryptr->classptr)
		classptr = (IRCD_CLASSENTRY *)entryptr->classptr;
	else
	{
		ret = class_find(&(_ircd_conf.classes), entryptr->classname, &classptr);
		if (ret < 0) /* hrm...*/
		{
			if (err)
				*err = ret;
			return -2;
		}
		entryptr->classptr = (void *)classptr;
	}

    if (!(classptr->flags & CLASS_FLAGS_BYPASSLIMITS))
	{
	    if (classptr->num_on >= classptr->maxclients)
	        return -3;

		if (classptr->vlimit && (classptr->num_on > classptr->vlimit))
		{
			IRCD_IPENTRY	*ipptr = ipentry_find(NULL, pptr->c_rem_ipaddr);
			if (ipptr && ((ipptr->local_connect_num+1) >= classptr->vlimit))
				return -6;
		}

	    if (classptr->uhlimit || classptr->hlimit || classptr->uvlimit)
	    {
			IRCD_CLIENT		*clptr;
			int				sameh = 0;
			int				sameuh = 0;
			int				sameuv = 0;
			ircd_uint32_t	vifaddr;
			
			if ((classptr->num_on >= classptr->uhlimit) ||
				(classptr->num_on >= classptr->hlimit) ||
				(classptr->num_on >= classptr->uvlimit))
			{
				vifaddr = pptr->c_rem_ipaddr & 0xffffff00;

				for(clptr=classptr->clients;clptr;clptr=clptr->c_cl_next)
				{
					if (clptr->c_rem_ipaddr == pptr->c_rem_ipaddr)
					{
						if (classptr->hlimit && (++sameh >= classptr->hlimit))
							return -4;
						if (classptr->uhlimit &&
								!strcasecmp(clptr->username, pptr->username) &&
								(++sameuh >= classptr->uhlimit))
							return -5;
					}
					if (classptr->uvlimit &&
							((clptr->c_rem_ipaddr & 0xffffff00) == vifaddr))
					{
						if (!strcasecmp(clptr->username, pptr->username) &&
								(++sameuv >= classptr->uvlimit))
							return -7;
					}
				}
			}
		}
	}
	if (classptr->flags & CLASS_FLAGS_HIDEHOSTNAME)
	{
		if (*classptr->hostname)
			strcpy(pptr->hostname, classptr->hostname);
		if (!*classptr->hostname ||
			(classptr->hostname[strlen(classptr->hostname)-1] == '.'))
		{
			support_strncpy(pptr->hostname+strlen(classptr->hostname),
				meptr->name, sizeof(pptr->hostname)-strlen(classptr->hostname));
		}
		strcpy(pptr->c_ipstring, "127.0.0.1");
	}
	add_client_to_cllist(&(classptr->clients), pptr);
	classptr->num_on++;
	pptr->c_class = classptr;
	pptr->c_maxsendq = classptr->maxsendq;
	pptr->c_pingfreq = classptr->pingfreq;
	pptr->c_rate_msg_limit = classptr->floodnum * 3;
	pptr->c_rate_time_limit = classptr->floodtime;
	pptr->c_rate_exit_num_limit = classptr->floodexit;
	return 0;
}

int motd_unload(IRCD_MOTD **motd)
{
	if (!motd || !*motd)
		return EINVAL;
	if ((*motd)->lines != NULL)
	{
		char **line;

		for(line=(*motd)->lines;*line;line++)
			MyFree(*line);
		MyFree((*motd)->lines);
		MyFree(*motd);
		*motd = NULL;
	}
	return 0;
}

int motd_load(char *filename, IRCD_MOTD **motd_ptr)
{
	IRCD_MOTD	*motd_new;
	aFileBuf	*fb;
	int			ret;
	int			numlines = 0;
	char		line[128];
	char		*ptr;
	struct stat	sb;

	if ((filename == NULL) || !*filename || (motd_ptr == NULL))
		return -EINVAL;

	motd_new = (IRCD_MOTD *)MyMalloc(sizeof(IRCD_MOTD));
	if (motd_new == NULL)
		return -ENOMEM;
	motd_new->size = 50;
	motd_new->lines = (char **)MyMalloc(sizeof(char *)*motd_new->size);
	if (motd_new->lines == NULL)
	{
		motd_unload(&motd_new);
		return -ENOMEM;
	}
	fb = fbopen(filename, "r");
	if (fb == NULL)
	{
		ret = errno;
		motd_unload(&motd_new);
		return ret ? -ret : -1;
	}
	if (fbstat(&sb, fb) == -1)
	{
		ret = errno;
		motd_unload(&motd_new);
		return ret ? -ret : -1;
	}
	motd_new->motd_time = *localtime(&(sb.st_mtime));
	while(fbgets(line, sizeof(line)-1, fb) != NULL)
	{
		if ((ptr = strchr(line, '\n')) != NULL)
			*ptr = '\0';
		if (++numlines == motd_new->size)
		{
			motd_new->size += 25;
			motd_new->lines = (char **)MyRealloc(motd_new->lines,
				sizeof(char *)*motd_new->size);
			if (motd_new->lines == NULL)
			{
				fbclose(fb);
				motd_unload(&motd_new);
				return -ENOMEM;
			}
		}
		if ((motd_new->lines[numlines-1] = MyStrdup(line)) == NULL)
		{
			fbclose(fb);
			motd_unload(&motd_new);
			return -ENOMEM;
		}
	}
	motd_new->lines[numlines] = NULL;
	fbclose(fb);
	*motd_ptr = motd_new;
	return 0;
}

char *conf_myname_for_link(char *name, int hostmask, char *buf)
{
	char *start = name;

	if ((hostmask <= 0) || (hostmask > 5))
		return name;
	while(hostmask-- && name)
	{
		name++;
		name = strchr(name, '.');
	}
	if (!name)
		return start;
	buf[0] = '*';
	strcpy(buf+1, name);
	return buf;
}

void logs_init(void)
{
	char	*logname;

	if (vars_value_get_bynum(IRCD_VAR_GLINE_LOG) &&
		((logname = (char *)vars_value_get_bynum(IRCD_VAR_GLINE_LOGFILE)) != NULL))
	{
		if (!_ircd_conf.log_gline ||
				!log_samename(_ircd_conf.log_gline, logname))
		{
			log_close(_ircd_conf.log_gline);
			_ircd_conf.log_gline = log_open(logname);
		}
	}

	if (vars_value_get_bynum(IRCD_VAR_BADOPER_LOG) &&
		((logname = (char *)vars_value_get_bynum(IRCD_VAR_BADOPER_LOGFILE)) != NULL))
	{
		if (!_ircd_conf.log_badoper ||
				!log_samename(_ircd_conf.log_badoper, logname))
		{
			log_close(_ircd_conf.log_badoper);
			_ircd_conf.log_badoper = log_open(logname);
		}
	}

	if (vars_value_get_bynum(IRCD_VAR_GOODOPER_LOG) &&
			((logname = (char *)vars_value_get_bynum(IRCD_VAR_GOODOPER_LOGFILE)) != NULL))
	{
		if (!_ircd_conf.log_goodoper ||
				!log_samename(_ircd_conf.log_goodoper, logname))
		{
			log_close(_ircd_conf.log_goodoper);
			_ircd_conf.log_goodoper = log_open(logname);
		}
	}

	if (vars_value_get_bynum(IRCD_VAR_CLONES_LOG) &&
			((logname = (char *)vars_value_get_bynum(IRCD_VAR_CLONES_LOGFILE)) != NULL))
	{
		if (!_ircd_conf.log_clones ||
				!log_samename(_ircd_conf.log_clones, logname))
		{
			log_close(_ircd_conf.log_clones);
			_ircd_conf.log_clones = log_open(logname);
		}
	}

	if (vars_value_get_bynum(IRCD_VAR_USERS_LOG) &&
			((logname = (char *)vars_value_get_bynum(IRCD_VAR_USERS_LOGFILE)) != NULL))
	{
		if (!_ircd_conf.log_users ||
				!log_samename(_ircd_conf.log_users, logname))
		{
			log_close(_ircd_conf.log_users);
			_ircd_conf.log_users = log_open(logname);
		}
	}
}

static void logs_flushall(void)
{
	log_flush(_ircd_conf.log_gline);
	log_flush(_ircd_conf.log_dline);
	log_flush(_ircd_conf.log_badoper);
	log_flush(_ircd_conf.log_goodoper);
	log_flush(_ircd_conf.log_clones);
	log_flush(_ircd_conf.log_users);
}

#if 0
static void logs_closeall(void)
{
	if (_ircd_conf.log_gline)
	{
		log_close(_ircd_conf.log_gline);
		_ircd_conf.log_gline = NULL;
	}
	if (_ircd_conf.log_dline)
	{
		log_close(_ircd_conf.log_dline);
		_ircd_conf.log_dline = NULL;
	}
	if (_ircd_conf.log_badoper)
	{
		log_close(_ircd_conf.log_badoper);
		_ircd_conf.log_badoper = NULL;
	}
	if (_ircd_conf.log_goodoper)
	{
		log_close(_ircd_conf.log_goodoper);
		_ircd_conf.log_goodoper = NULL;
	}
	if (_ircd_conf.log_clones)
	{
		log_close(_ircd_conf.log_clones);
		_ircd_conf.log_clones = NULL;
	}
	if (_ircd_conf.log_users)
	{
		log_close(_ircd_conf.log_users);
		_ircd_conf.log_users = NULL;
	}
}
#endif

void stringdeny_add(STRINGDENY **stlist, char *name, char *reason, int expires, u_int flags)
{
	STRINGDENY	 *sptr;

	if (!stlist || !name || !reason)
		return;

	sptr = (STRINGDENY *)MyMalloc(sizeof(STRINGDENY));
	sptr->name = MyStrdup(name);
	sptr->reason = MyStrdup(reason);
	sptr->expires = expires;
	sptr->flags = flags;
	sptr->prev = NULL;
	if ((sptr->next = *stlist) != NULL)
		(*stlist)->prev = sptr; 
	*stlist = sptr; 
}

void stringdeny_del(STRINGDENY **stlist, STRINGDENY *sptr)
{
	if (sptr->prev)
		sptr->prev->next = sptr->next;
	else	
		*stlist = sptr->next;
	if (sptr->next)
		sptr->next->prev = sptr; 
	if (sptr->name)
		MyFree(sptr->name);
	MyFree(sptr);
}

STRINGDENY *stringdeny_find(STRINGDENY **stlist, char *name, int expires)
{
	STRINGDENY	*sptr;
	STRINGDENY	*nextsptr;

	if (!name)
		return NULL;

	for(sptr=*stlist;sptr;sptr=nextsptr)
	{
		nextsptr = sptr->next;
		if ((expires > 0) && (sptr->expires > 0) &&
						(expires > sptr->expires))
		{
			stringdeny_del(stlist, sptr);
			continue;
		}
		if (sptr->name && !mycmp(sptr->name, name))
			return sptr;
	}
	return NULL;
}

STRINGDENY *stringdeny_match(STRINGDENY **stlist, char *name, int expires)
{
	STRINGDENY	*sptr;
	STRINGDENY	*nextsptr;

	if (!name)
		return NULL;

	for(sptr=*stlist;sptr;sptr=nextsptr)
	{
		nextsptr = sptr->next;
		if ((expires > 0) && (sptr->expires > 0) &&
						(expires > sptr->expires))
		{
			stringdeny_del(stlist, sptr);
			continue;
		}
		if (sptr->name && !match(sptr->name, name))
			return sptr;
	}
	return NULL;
}

void stringdeny_free(STRINGDENY **stlist)
{
	STRINGDENY *n;									  
															 
	while(*stlist)
	{   
		n = (*stlist)->next;
		if ((*stlist)->name)
			MyFree((*stlist)->name);
		if ((*stlist)->reason)
			MyFree((*stlist)->reason);
		MyFree(*stlist);
		*stlist = n;
	}
}

static void _isupport_add_num(IRCD_STRINGLIST **sl, char *buffer, u_int bufsize, char **curptr, char *variable, int num)
{
	char			string[128];
	u_int			vlen = strlen(variable);
	u_int			slen;
	u_int			tlen;

	sprintf(string, "%d", num);
	slen = strlen(string);
	tlen = vlen + slen + 1;

	if (tlen + 1 > bufsize)
		return;	/* doh, too big for buffer */

	if (((*curptr) + tlen + 1) >= (buffer + bufsize))
	{
		stringlist_add(sl, buffer);
		*curptr = buffer;
	}

	if (*curptr != buffer)
		*((*curptr)++) = ' ';
	strcpy(*curptr, variable);
	*((*curptr) + vlen) = '=';
	strcpy((*curptr) + vlen + 1, string);
	(*curptr) += tlen;
}

static void _isupport_add_string(IRCD_STRINGLIST **sl, char *buffer, u_int bufsize, char **curptr, char *variable, char *string)
{
	u_int			vlen = strlen(variable);
	u_int			slen = strlen(string);
	u_int			tlen;

	tlen = vlen;
	if (slen)
		tlen += slen + 1;

	if (tlen + 1 > bufsize)
		return;	/* doh, too big for buffer */

	if ((*curptr + tlen + 1) >= (buffer + bufsize))
	{
		stringlist_add(sl, buffer);
		*curptr = buffer;
	}

	if (*curptr != buffer)
		*((*curptr)++) = ' ';
	strcpy(*curptr, variable);
	if (slen)
	{
		*((*curptr) + vlen) = '=';
		strcpy((*curptr) + vlen + 1, string);
	}
	(*curptr) += tlen;
}

static void _isupport_done(IRCD_STRINGLIST **sl, char *buffer)
{
	stringlist_add(sl, buffer);
}

static int _deny_acl_add(IRCD_DENY_ACL **acl_list, IRCD_DENY_ACL **acl_last, char *oper, char *server, char *mask, u_int flags, IRCD_DENY_ACL **aclret)
{
	char			*oper_username;
	char			*oper_hostname;
	char			*mask_username;
	char			*mask_hostname;
	IRCD_DENY_ACL	*acl;
	char			tempoper[256];
	char			tempmask[256];

	support_strncpy(tempoper, oper, sizeof(tempoper));
	support_strncpy(tempmask, mask, sizeof(tempmask));

	ircd_userhost_split(tempoper, &oper_username, &oper_hostname);
	ircd_userhost_split(tempmask, &mask_username, &mask_hostname);

	for(acl=*acl_list;acl;acl=acl->next)
	{
		if (!match(acl->src_oper_username, oper_username) &&
				!match(acl->src_oper_hostname, oper_hostname) &&
				!match(acl->src_server, server) &&
				!match(acl->dest_mask_username, mask_username) &&
				!match(acl->dest_mask_hostname, mask_hostname) &&
				(acl->flags == flags))
		{
			*aclret = acl;
			return EEXIST;
		}
	}

	acl = (IRCD_DENY_ACL *)MyMalloc(sizeof(IRCD_DENY_ACL));
	acl->flags = flags;
	acl->src_oper_username = MyStrdup(oper_username);
	acl->src_oper_hostname = MyStrdup(oper_hostname);
	acl->src_server = MyStrdup(server);
	acl->dest_mask_username = MyStrdup(mask_username);
	acl->dest_mask_hostname = MyStrdup(mask_hostname);
	acl->next = NULL;

	if (*acl_last)
		(*acl_last)->next = acl;
	else
		*acl_list = acl;

	*acl_last = acl;
	*aclret = acl;
	return 0;
}

static void _deny_acls_free(IRCD_DENY_ACL **acl_list, IRCD_DENY_ACL **acl_last)
{
	IRCD_DENY_ACL	*n;

	while(*acl_list)
	{
		n = (*acl_list)->next;

		MyFree((*acl_list)->src_oper_username);
		MyFree((*acl_list)->src_oper_hostname);
		MyFree((*acl_list)->src_server);
		MyFree((*acl_list)->dest_mask_username);
		MyFree((*acl_list)->dest_mask_hostname);
		MyFree(*acl_list);

		*acl_list = n;
	}
	*acl_last = NULL;
}

void build_isupport(void)
{
	char	buf[400];
	char	*curptr = buf;
	int		len;

	stringlist_free(&CSIrcd_ISupport);
	_isupport_add_string(&CSIrcd_ISupport, buf, sizeof(buf), &curptr,
			"NETWORK", NETWORK_NAME);
	_isupport_add_string(&CSIrcd_ISupport, buf, sizeof(buf), &curptr,
			"STD", "i-d");
	_isupport_add_string(&CSIrcd_ISupport, buf, sizeof(buf), &curptr,
			"STATUSMSG", "@+");
	_isupport_add_string(&CSIrcd_ISupport, buf, sizeof(buf), &curptr,
			"PREFIX", "(ov)@+");
	_isupport_add_string(&CSIrcd_ISupport, buf, sizeof(buf), &curptr,
			"CHANTYPES", "#&");
	_isupport_add_num(&CSIrcd_ISupport, buf, sizeof(buf), &curptr,
			"NICKLEN", NICKLEN);
	len = (int)vars_value_get_bynum(IRCD_VAR_MAX_TOPIC_LEN);
	_isupport_add_num(&CSIrcd_ISupport, buf, sizeof(buf), &curptr,
			"TOPICLEN", len);
	len = (int)vars_value_get_bynum(IRCD_VAR_MAX_KICK_REASON_LEN);
	_isupport_add_num(&CSIrcd_ISupport, buf, sizeof(buf), &curptr,
			"KICKLEN", len);
	_isupport_add_string(&CSIrcd_ISupport, buf, sizeof(buf), &curptr,
			"CHANMODES", "b,k,l,imnpst");
	len = (int)vars_value_get_bynum(IRCD_VAR_MAX_MODE_PARAMS);
	_isupport_add_num(&CSIrcd_ISupport, buf, sizeof(buf), &curptr,
			"MODES", len);
	len = (int)vars_value_get_bynum(IRCD_VAR_MAX_BANS);
	_isupport_add_num(&CSIrcd_ISupport, buf, sizeof(buf), &curptr,
			"MAXBANS", len);
	_isupport_add_string(&CSIrcd_ISupport, buf, sizeof(buf), &curptr,
			"CASEMAPPING", "rfc1459");
	_isupport_add_string(&CSIrcd_ISupport, buf, sizeof(buf), &curptr,
			"CHARSET", "ascii");

	/* must be last right now */
	_isupport_add_num(&CSIrcd_ISupport, buf, sizeof(buf), &curptr,
			"MAXCHANNELS", MAXCHANNELSPERUSER);
	_isupport_done(&CSIrcd_ISupport, buf);
}

void send_isupport_to_one(IRCD_CLIENT *one)
{
	IRCD_STRINGLIST		*sl;
	char				*ptr;

	if (IsClient(one))
	{
		int maxchan = (int)vars_value_get_bynum(IRCD_VAR_DEFAULT_MAXCHANNELS);
		if (MyClient(one))
		{
			if (one->c_class->maxchannels >= 0)
				maxchan = one->c_class->maxchannels;
		}

		for(sl=CSIrcd_ISupport;sl;sl=sl->next)
		{
			ptr = strstr(sl->string, "MAXCHANNELS=");
			if (ptr)
			{
				ptr += 12;
				sprintf(ptr, "%d", maxchan);
			}
			send_prefix_to_client(meptr, one, IRCD_REPLY(RPL_ISUPPORT,
				one->name), sl->string);	
		}
	}
}

IRCD_DENY_ACL *deny_acl_find(IRCD_DENY_ACL *acl, char *oper_username, char *oper_hostname, char *server, char *mask_username, char *mask_hostname)
{
	for(;acl;acl=acl->next)
	{
		if ((!oper_username ||
				!match(acl->src_oper_username, oper_username)) &&
			(!oper_hostname ||
				!match(acl->src_oper_hostname, oper_hostname)) &&
			(!server ||
				!match(acl->src_server, server)) &&
			(!mask_username ||
				!match(acl->dest_mask_username, mask_username)) &&
			(!mask_username ||
				!match(acl->dest_mask_hostname, mask_hostname)))
		{
			return acl;
		}
	}
	return NULL;
}

